import time
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import Optional
from typing import Type

from ddtrace import config as tracer_config
from ddtrace.debugging._config import config
from ddtrace.debugging._expressions import dd_compile
from ddtrace.debugging._probe.model import CaptureLimits
from ddtrace.debugging._probe.model import DDExpression
from ddtrace.debugging._probe.model import DEFAULT_PROBE_CONDITION_ERROR_RATE
from ddtrace.debugging._probe.model import DEFAULT_PROBE_RATE
from ddtrace.debugging._probe.model import DEFAULT_SNAPSHOT_PROBE_RATE
from ddtrace.debugging._probe.model import ExpressionTemplateSegment
from ddtrace.debugging._probe.model import LiteralTemplateSegment
from ddtrace.debugging._probe.model import LogFunctionProbe
from ddtrace.debugging._probe.model import LogLineProbe
from ddtrace.debugging._probe.model import MetricFunctionProbe
from ddtrace.debugging._probe.model import MetricLineProbe
from ddtrace.debugging._probe.model import Probe
from ddtrace.debugging._probe.model import ProbeType
from ddtrace.internal.logger import get_logger
from ddtrace.internal.remoteconfig.client import ConfigMetadata
from ddtrace.internal.remoteconfig.client import RemoteConfigCallBack
from ddtrace.internal.utils.cache import LFUCache


log = get_logger(__name__)


_EXPRESSION_CACHE = LFUCache()


def xlate_keys(d, mapping):
    # type: (Dict[str, Any], Dict[str, str]) -> Dict[str, Any]
    return {mapping.get(k, k): v for k, v in d.items()}


def _invalid_expression(_):
    """Forces probes with invalid expression/conditions to never trigger.

    Any signs of invalid conditions in logs is an indication of a problem with
    the expression compiler.
    """
    return None


INVALID_EXPRESSION = _invalid_expression


def _compile_expression(expr):
    # type: (Optional[Dict[str, Any]]) -> Optional[DDExpression]
    global _EXPRESSION_CACHE, INVALID_EXPRESSION

    if expr is None:
        return None

    ast = expr["json"]

    def compile_or_invalid(expr):
        # type: (str) -> Callable[[Dict[str, Any]], Any]
        try:
            return dd_compile(ast)
        except Exception:
            log.error("Cannot compile expression: %s", expr, exc_info=True)
            return INVALID_EXPRESSION

    dsl = expr["dsl"]

    compiled = _EXPRESSION_CACHE.get(dsl, compile_or_invalid)  # type: Callable[[Dict[str, Any]], Any]

    if compiled is INVALID_EXPRESSION:
        log.error("Cannot compile expression: %s", dsl, exc_info=True)

    return DDExpression(dsl=dsl, callable=compiled)


def _compile_segment(segment):
    if segment.get("str", ""):
        return LiteralTemplateSegment(str_value=segment["str"])
    elif segment.get("json", None) is not None:
        return ExpressionTemplateSegment(expr=_compile_expression(segment))

    # what type of error we should show here?
    return None


def _match_env_and_version(probe):
    # type: (Probe) -> bool
    probe_version = probe.tags.get("version", None)
    probe_env = probe.tags.get("env", None)

    return (probe_version is None or probe_version == tracer_config.version) and (
        probe_env is None or probe_env == tracer_config.env
    )


def _filter_by_env_and_version(f):
    # type: (Callable[..., Iterable[Probe]]) -> Callable[..., Iterable[Probe]]
    def _wrapper(*args, **kwargs):
        # type: (Any, Any) -> Iterable[Probe]
        return [_ for _ in f(*args, **kwargs) if _match_env_and_version(_)]

    return _wrapper


def _create_probe_based_on_location(args, attribs, line_class, function_class):
    # type: (Dict[str, Any], Dict[str, Any], Type, Type) -> Any
    if attribs["where"].get("sourceFile", None):
        ProbeType = line_class
        args["source_file"] = attribs["where"]["sourceFile"]
        args["line"] = int(attribs["where"]["lines"][0])
    else:
        ProbeType = function_class
        args["module"] = attribs["where"].get("type") or attribs["where"]["typeName"]
        args["func_qname"] = attribs["where"].get("method") or attribs["where"]["methodName"]
        args["evaluate_at"] = attribs.get("evaluateAt")

    return ProbeType(**args)


def probe_factory(attribs):
    # type: (Dict[str, Any]) -> Probe
    """
    Create a new Probe instance.
    """
    try:
        _type = attribs["type"]
        _id = attribs["id"]
    except KeyError as e:
        raise ValueError("Invalid probe attributes: %s" % e)

    if _type == ProbeType.LOG_PROBE:
        take_snapshot = attribs.get("captureSnapshot", False)

        rate = DEFAULT_SNAPSHOT_PROBE_RATE if take_snapshot else DEFAULT_PROBE_RATE
        sampling = attribs.get("sampling")
        if sampling is not None:
            rate = sampling.get("snapshotsPerSecond", rate)

        args = dict(
            probe_id=_id,
            condition=_compile_expression(attribs.get("when")),
            tags=dict(_.split(":", 1) for _ in attribs.get("tags", [])),
            rate=rate,
            limits=CaptureLimits(
                **xlate_keys(
                    attribs["capture"],
                    {
                        "maxReferenceDepth": "max_level",
                        "maxCollectionSize": "max_size",
                        "maxLength": "max_len",
                        "maxFieldDepth": "max_fields",
                    },
                )
            )
            if "capture" in attribs
            else None,
            condition_error_rate=DEFAULT_PROBE_CONDITION_ERROR_RATE,  # TODO: should we take rate limit out of Probe?
            take_snapshot=take_snapshot,
            template=attribs.get("template"),
            segments=[_compile_segment(segment) for segment in attribs.get("segments", [])],
        )

        return _create_probe_based_on_location(args, attribs, LogLineProbe, LogFunctionProbe)

    elif _type == ProbeType.METRIC_PROBE:
        args = dict(
            probe_id=_id,
            condition=_compile_expression(attribs.get("when")),
            tags=dict(_.split(":", 1) for _ in attribs.get("tags", [])),
            name=attribs["metricName"],
            kind=attribs["kind"],
            rate=DEFAULT_PROBE_RATE,  # unused
            condition_error_rate=DEFAULT_PROBE_CONDITION_ERROR_RATE,  # TODO: should we take rate limit out of Probe?
            value=_compile_expression(attribs.get("value")),
        )

        return _create_probe_based_on_location(args, attribs, MetricLineProbe, MetricFunctionProbe)

    raise ValueError("Unknown probe type: %s" % _type)


@_filter_by_env_and_version
def get_probes(config_id, config):
    # type: (str, dict) -> Iterable[Probe]
    return [probe_factory(config)]


log = get_logger(__name__)


class ProbePollerEvent(object):
    NEW_PROBES = 0
    DELETED_PROBES = 1
    MODIFIED_PROBES = 2
    STATUS_UPDATE = 3


ProbePollerEventType = int


class ProbeRCAdapter(RemoteConfigCallBack):
    """Probe configuration adapter for the RCM client.

    This adapter turns configuration events from the RCM client into probe
    events that can be handled easily by the debugger.
    """

    def __init__(self, callback):
        # type: (Callable[[ProbePollerEventType, Iterable[Probe]], None]) -> None
        self._callback = callback
        self._configs = {}  # type: Dict[str, Dict[str, Probe]]
        self._next_status_update_timestamp()

    def _next_status_update_timestamp(self):
        # type: () -> None
        self._status_timestamp = time.time() + config.diagnostics_interval

    def _dispatch_probe_events(self, prev_probes, next_probes):
        new_probes = [p for _, p in next_probes.items() if _ not in prev_probes]
        deleted_probes = [p for _, p in prev_probes.items() if _ not in next_probes]
        modified_probes = []  # DEV: Probes are currently immutable

        if deleted_probes:
            self._callback(ProbePollerEvent.DELETED_PROBES, deleted_probes)
        if modified_probes:
            self._callback(ProbePollerEvent.MODIFIED_PROBES, modified_probes)
        if new_probes:
            self._callback(ProbePollerEvent.NEW_PROBES, new_probes)

    def _update_probes_for_config(self, config_id, config):
        # type: (str, Any) -> None
        prev_probes = self._configs.get(config_id, {})  # type: Dict[str, Probe]
        next_probes = (
            {probe.probe_id: probe for probe in get_probes(config_id, config)} if config not in (None, False) else {}
        )  # type: Dict[str, Probe]

        self._dispatch_probe_events(prev_probes, next_probes)

        if next_probes:
            self._configs[config_id] = next_probes
        else:
            self._configs.pop(config_id, None)

    def __call__(self, metadata, config):
        # type: (Optional[ConfigMetadata], Any) -> None

        # DEV: We emit a status update event here to avoid having to spawn a
        # separate thread for this.
        if time.time() > self._status_timestamp:
            log.debug("Emitting probe status log messages")
            probes = [probe for config in self._configs.values() for probe in config.values()]
            self._callback(ProbePollerEvent.STATUS_UPDATE, probes)
            self._next_status_update_timestamp()

        if metadata is None:
            log.warning("no metadata was provided")
            return

        self._update_probes_for_config(metadata.id, config)



import httpx
import pytest
import six

from ddtrace import config
from ddtrace.contrib.httpx.patch import HTTPX_VERSION
from ddtrace.contrib.httpx.patch import patch
from ddtrace.contrib.httpx.patch import unpatch
from ddtrace.pin import Pin
from ddtrace.settings.http import HttpConfig
from ddtrace.vendor.wrapt import ObjectProxy
from tests.utils import override_config
from tests.utils import override_http_config


# host:port of httpbin_local container
HOST = "localhost"
PORT = 8001

DEFAULT_HEADERS = {
    "User-Agent": "python-httpx/x.xx.x",
}


pytestmark = pytest.mark.skipif(HTTPX_VERSION >= (0, 11), reason="httpx<=0.10 Client is asynchronous")


def get_url(path):
    # type: (str) -> str
    return "http://{}:{}{}".format(HOST, PORT, path)


@pytest.fixture(autouse=True)
def patch_httpx():
    patch()
    try:
        yield
    finally:
        unpatch()


def test_patching():
    """
    When patching httpx library
        We wrap the correct methods
    When unpatching httpx library
        We unwrap the correct methods
    """
    assert isinstance(httpx.Client.send, ObjectProxy)
    unpatch()
    assert not isinstance(httpx.Client.send, ObjectProxy)


@pytest.mark.asyncio
async def test_httpx_service_name(tracer, test_spans):
    """
    When using split_by_domain
        We set the span service name as a text type and not binary
    """
    client = httpx.Client()
    Pin.override(client, tracer=tracer)

    with override_config("httpx", {"split_by_domain": True}):
        resp = await client.get(get_url("/status/200"))
    assert resp.status_code == 200

    traces = test_spans.pop_traces()
    assert len(traces) == 1

    spans = traces[0]
    assert len(spans) == 1
    assert isinstance(spans[0].service, six.text_type)


@pytest.mark.asyncio
async def test_get_200(snapshot_context):
    token = "tests.contrib.httpx.test_httpx.test_get_200"
    url = get_url("/status/200")

    with snapshot_context(token=token):
        resp = await httpx.get(url, headers=DEFAULT_HEADERS)
        assert resp.status_code == 200


@pytest.mark.asyncio
async def test_configure_service_name(snapshot_context):
    """
    When setting ddtrace.config.httpx.service_name directly
        We use the value from ddtrace.config.httpx.service_name
    """
    token = "tests.contrib.httpx.test_httpx.test_configure_service_name"
    url = get_url("/status/200")

    with override_config("httpx", {"service_name": "test-httpx-service-name"}):
        with snapshot_context(token=token):
            resp = await httpx.get(url, headers=DEFAULT_HEADERS)
            assert resp.status_code == 200


@pytest.mark.asyncio
async def test_configure_service_name_pin(tracer, test_spans):
    """
    When setting service name via a Pin
        We use the value from the Pin
    """
    url = get_url("/status/200")

    def assert_spans(test_spans, service):
        test_spans.assert_trace_count(1)
        test_spans.assert_span_count(1)
        assert test_spans.spans[0].service == service
        test_spans.reset()

    # override the tracer on the default sync client
    # DEV: `httpx.get` will call `with Client() as client: client.get()`
    Pin.override(httpx.Client, tracer=tracer)

    # sync client
    client = httpx.Client()
    Pin.override(client, service="sync-client", tracer=tracer)

    resp = await httpx.get(url, headers=DEFAULT_HEADERS)
    assert resp.status_code == 200
    assert_spans(test_spans, service=None)

    resp = await client.get(url, headers=DEFAULT_HEADERS)
    assert resp.status_code == 200
    assert_spans(test_spans, service="sync-client")


@pytest.mark.subprocess(
    env=dict(
        DD_HTTPX_SERVICE="env-overridden-service-name",
        DD_SERVICE="global-service-name",
    )
)
def test_configure_service_name_env():
    """
    When setting DD_HTTPX_SERVICE env variable
        When DD_SERVICE is also set
            We use the value from DD_HTTPX_SERVICE
    """
    import asyncio
    import sys

    import httpx

    from ddtrace.contrib.httpx import patch
    from tests.contrib.httpx.test_httpx import get_url
    from tests.utils import snapshot_context

    patch()
    url = get_url("/status/200")

    async def test():
        token = "tests.contrib.httpx.test_httpx.test_configure_service_name_env"
        with snapshot_context(wait_for_num_traces=1, token=token):
            DEFAULT_HEADERS = {
                "User-Agent": "python-httpx/x.xx.x",
            }
            await httpx.get(url, headers=DEFAULT_HEADERS)

    if sys.version_info >= (3, 7, 0):
        asyncio.run(test())
    else:
        asyncio.get_event_loop().run_until_complete(test())


@pytest.mark.subprocess(env=dict(DD_SERVICE="global-service-name"))
def test_configure_global_service_name_env():
    """
    When only setting DD_SERVICE
        We use the value from DD_SERVICE for the service name
    """

    import asyncio
    import sys

    import httpx

    from ddtrace.contrib.httpx import patch
    from tests.contrib.httpx.test_httpx import get_url
    from tests.utils import snapshot_context

    patch()
    url = get_url("/status/200")

    async def test():
        token = "tests.contrib.httpx.test_httpx.test_configure_global_service_name_env"
        with snapshot_context(wait_for_num_traces=1, token=token):
            DEFAULT_HEADERS = {
                "User-Agent": "python-httpx/x.xx.x",
            }
            await httpx.get(url, headers=DEFAULT_HEADERS)

    if sys.version_info >= (3, 7, 0):
        asyncio.run(test())
    else:
        asyncio.get_event_loop().run_until_complete(test())


@pytest.mark.asyncio
async def test_get_500(snapshot_context):
    """
    When the status code is 500
        We mark the span as an error
    """
    token = "tests.contrib.httpx.test_httpx.test_get_500"
    url = get_url("/status/500")
    with snapshot_context(wait_for_num_traces=1, token=token):
        resp = await httpx.get(url, headers=DEFAULT_HEADERS)
        assert resp.status_code == 500


@pytest.mark.asyncio
async def test_split_by_domain(snapshot_context):
    """
    When split_by_domain is configured
        We set the service name to the <host>:<port>
    """
    token = "tests.contrib.httpx.test_httpx.test_split_by_domain"
    url = get_url("/status/200")

    with override_config("httpx", {"split_by_domain": True}):
        with snapshot_context(wait_for_num_traces=1, token=token):
            resp = await httpx.get(url, headers=DEFAULT_HEADERS)
            assert resp.status_code == 200


@pytest.mark.asyncio
async def test_trace_query_string(snapshot_context):
    """
    When trace_query_string is enabled
        We include the query string as a tag on the span
    """
    token = "tests.contrib.httpx.test_httpx.test_trace_query_string"
    url = get_url("/status/200?some=query&string=args")
    # Caveat(avara1986): Docker image "httpbin" set as user-agent "python-httpx/0.23.0".
    # your local container or the CI container, could have different version. We set
    # the user agent to use the same version.
    headers = {
        "User-Agent": "python-httpx/x.xx.x",
    }
    with override_http_config("httpx", {"trace_query_string": True}):
        with snapshot_context(wait_for_num_traces=1, token=token):
            resp = await httpx.get(url, headers=headers)
            assert resp.status_code == 200


@pytest.mark.asyncio
async def test_request_headers(snapshot_context):
    """
    When request headers are configured for this integration
        We add the request headers as tags on the span
    """
    token = "tests.contrib.httpx.test_httpx.test_request_headers"
    url = get_url("/response-headers?Some-Response-Header=Response-Value")

    headers = {
        "Some-Request-Header": "Request-Value",
        "User-Agent": "python-httpx/x.xx.x",
    }

    try:
        config.httpx.http.trace_headers(["Some-Request-Header", "Some-Response-Header"])
        with snapshot_context(wait_for_num_traces=1, token=token):
            resp = await httpx.get(url, headers=headers)
            assert resp.status_code == 200
    finally:
        config.httpx.http = HttpConfig()


@pytest.mark.asyncio
async def test_distributed_tracing_headers():
    """
    By default
        Distributed tracing headers are added to outbound requests
    """
    url = get_url("/headers")

    def assert_request_headers(response):
        data = response.json()
        assert "X-Datadog-Trace-Id" in data["headers"]
        assert "X-Datadog-Parent-Id" in data["headers"]
        assert "X-Datadog-Sampling-Priority" in data["headers"]

    resp = await httpx.get(url, headers=DEFAULT_HEADERS)
    assert_request_headers(resp)


@pytest.mark.asyncio
async def test_distributed_tracing_disabled():
    """
    When distributed_tracing is disabled
        We do not add distributed tracing headers to outbound requests
    """
    url = get_url("/headers")

    def assert_request_headers(response):
        data = response.json()
        assert "X-Datadog-Trace-Id" not in data["headers"]
        assert "X-Datadog-Parent-Id" not in data["headers"]
        assert "X-Datadog-Sampling-Priority" not in data["headers"]

    with override_config("httpx", {"distributed_tracing": False}):
        resp = await httpx.get(url, headers=DEFAULT_HEADERS)
        assert_request_headers(resp)


@pytest.mark.subprocess(env=dict(DD_HTTPX_DISTRIBUTED_TRACING="false"))
def test_distributed_tracing_disabled_env():
    """
    When disabling distributed tracing via env variable
        We do not add distributed tracing headers to outbound requests
    """

    import asyncio
    import sys

    import httpx

    from ddtrace.contrib.httpx import patch
    from tests.contrib.httpx.test_httpx import get_url

    patch()
    url = get_url("/headers")

    async def test():
        def assert_request_headers(response):
            data = response.json()
            assert "X-Datadog-Trace-Id" not in data["headers"]
            assert "X-Datadog-Parent-Id" not in data["headers"]
            assert "X-Datadog-Sampling-Priority" not in data["headers"]

        DEFAULT_HEADERS = {
            "User-Agent": "python-httpx/x.xx.x",
        }
        resp = await httpx.get(url, headers=DEFAULT_HEADERS)
        assert_request_headers(resp)

    if sys.version_info >= (3, 7, 0):
        asyncio.run(test())
    else:
        asyncio.get_event_loop().run_until_complete(test())



from typing import TYPE_CHECKING
from typing import Union  # noqa

from ddtrace.internal.logger import get_logger
from ddtrace.vendor.sqlcommenter import generate_sql_comment as _generate_sql_comment

from ..internal.utils import get_argument_value
from ..internal.utils import set_argument_value
from ..settings import _config as dd_config
from ..settings._database_monitoring import dbm_config


if TYPE_CHECKING:
    from typing import Optional

    from ddtrace import Span

DBM_PARENT_SERVICE_NAME_KEY = "ddps"
DBM_DATABASE_SERVICE_NAME_KEY = "dddbs"
DBM_ENVIRONMENT_KEY = "dde"
DBM_VERSION_KEY = "ddpv"
DBM_TRACE_PARENT_KEY = "traceparent"
DBM_TRACE_INJECTED_TAG = "_dd.dbm_trace_injected"


log = get_logger(__name__)


def default_sql_injector(dbm_comment, sql_statement):
    # type: (str, Union[str, bytes]) -> Union[str, bytes]
    try:
        if isinstance(sql_statement, bytes):
            return dbm_comment.encode("utf-8", errors="strict") + sql_statement
        return dbm_comment + sql_statement
    except (TypeError, ValueError):
        log.warning(
            "Linking Database Monitoring profiles to spans is not supported for the following query type: %s. "
            "To disable this feature please set the following environment variable: "
            "DD_DBM_PROPAGATION_MODE=disabled",
            type(sql_statement),
        )
    return sql_statement


class _DBM_Propagator(object):
    def __init__(self, sql_pos, sql_kw, sql_injector=default_sql_injector):
        self.sql_pos = sql_pos
        self.sql_kw = sql_kw
        self.sql_injector = sql_injector

    def inject(self, dbspan, args, kwargs):
        dbm_comment = self._get_dbm_comment(dbspan)
        if dbm_comment is None:
            # injection_mode is disabled
            return args, kwargs

        original_sql_statement = get_argument_value(args, kwargs, self.sql_pos, self.sql_kw)
        # add dbm comment to original_sql_statement
        sql_with_dbm_tags = self.sql_injector(dbm_comment, original_sql_statement)
        # replace the original query or procedure with sql_with_dbm_tags
        args, kwargs = set_argument_value(args, kwargs, self.sql_pos, self.sql_kw, sql_with_dbm_tags)
        return args, kwargs

    def _get_dbm_comment(self, db_span):
        # type: (Span) -> Optional[str]
        """Generate DBM trace injection comment and updates span tags
        This method will set the ``_dd.dbm_trace_injected: "true"`` tag
        on ``db_span`` if the configured injection mode is ``"full"``.
        """
        if dbm_config.propagation_mode == "disabled":
            return None

        # set the following tags if DBM injection mode is full or service
        dbm_tags = {
            DBM_PARENT_SERVICE_NAME_KEY: dd_config.service,
            DBM_ENVIRONMENT_KEY: dd_config.env,
            DBM_VERSION_KEY: dd_config.version,
            DBM_DATABASE_SERVICE_NAME_KEY: db_span.service,
        }

        if dbm_config.propagation_mode == "full":
            db_span.set_tag_str(DBM_TRACE_INJECTED_TAG, "true")
            dbm_tags[DBM_TRACE_PARENT_KEY] = db_span.context._traceparent

        sql_comment = _generate_sql_comment(**dbm_tags)
        if sql_comment:
            # replace leading whitespace with trailing whitespace
            return sql_comment.strip() + " "
        return ""



import threading
from time import sleep

from tests.debugging.mocking import TestDebugger


def foo():
    concealed = "secret"  # noqa
    return


def test():
    sleep(0.5)

    foo()

    (snapshot,) = TestDebugger._instance.test_queue

    assert snapshot.probe.probe_id == "run_module_test"
    assert snapshot.frame.f_locals["concealed"] == "secret"

    print("OK")


threading.Thread(target=test).start()



import json
from typing import TYPE_CHECKING

import attr

from ddtrace.appsec.iast import oce
from ddtrace.constants import APPSEC_ORIGIN_VALUE
from ddtrace.constants import IAST_CONTEXT_KEY
from ddtrace.constants import IAST_ENABLED
from ddtrace.constants import IAST_JSON
from ddtrace.constants import MANUAL_KEEP_KEY
from ddtrace.constants import ORIGIN_KEY
from ddtrace.ext import SpanTypes
from ddtrace.internal import _context
from ddtrace.internal.logger import get_logger
from ddtrace.internal.processor import SpanProcessor


if TYPE_CHECKING:  # pragma: no cover
    from ddtrace.span import Span

log = get_logger(__name__)


@attr.s(eq=False)
class AppSecIastSpanProcessor(SpanProcessor):
    def on_span_start(self, span):
        # type: (Span) -> None
        if span.span_type != SpanTypes.WEB:
            return
        oce.acquire_request()

    def on_span_finish(self, span):
        # type: (Span) -> None
        """Report reported vulnerabilities.

        Span Tags:
            - `_dd.iast.json`: Only when one or more vulnerabilities have been detected will we include the custom tag.
            - `_dd.iast.enabled`: Set to 1 when IAST is enabled in a request. If a request is disabled
              (e.g. by sampling), then it is not set.
        """
        if span.span_type != SpanTypes.WEB:
            return

        span.set_metric(IAST_ENABLED, 1.0)

        data = _context.get_item(IAST_CONTEXT_KEY, span=span)

        if data:
            span.set_tag_str(IAST_JSON, json.dumps(attr.asdict(data)))

            span.set_tag(MANUAL_KEEP_KEY)
            if span.get_tag(ORIGIN_KEY) is None:
                span.set_tag_str(ORIGIN_KEY, APPSEC_ORIGIN_VALUE)

        oce.release_request()



"""
This app exists to replicate and report on failures and degraded behavior that can arise when using ddtrace with
gunicorn
"""
import json
import os

from ddtrace import tracer
from ddtrace.contrib.wsgi import DDWSGIMiddleware
from ddtrace.profiling import bootstrap
import ddtrace.profiling.auto  # noqa
from tests.webclient import PingFilter


if os.getenv("_DD_TEST_IMPORT_AUTO"):
    import ddtrace.auto  # noqa: F401  # isort: skip

tracer.configure(
    settings={
        "FILTERS": [PingFilter()],
    }
)

SCHEDULER_SENTINEL = -1
assert bootstrap.profiler._scheduler._last_export not in (None, SCHEDULER_SENTINEL)
bootstrap.profiler._scheduler._last_export = SCHEDULER_SENTINEL


def aggressive_shutdown():
    tracer.shutdown(timeout=1)
    if hasattr(bootstrap, "profiler"):
        bootstrap.profiler._scheduler.stop()
        bootstrap.profiler.stop()


def simple_app(environ, start_response):
    if environ["RAW_URI"] == "/shutdown":
        aggressive_shutdown()
        data = b"goodbye"
    else:
        payload = {
            "profiler": {"is_active": bootstrap.profiler._scheduler._last_export != SCHEDULER_SENTINEL},
        }
        data = json.dumps(payload).encode("utf-8")

    start_response("200 OK", [("Content-Type", "text/plain"), ("Content-Length", str(len(data)))])
    return iter([data])


app = DDWSGIMiddleware(simple_app)



import sys
from types import FunctionType
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import cast

from six import PY3


try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore[assignment]

import bytecode as b
from bytecode import Bytecode
from bytecode import Compare
from bytecode import CompilerFlags
from bytecode import Instr
from bytecode import Label

from .compat import PYTHON_VERSION_INFO as PY


class WrappedFunction(Protocol):
    """A wrapped function."""

    __dd_wrapped__ = None  # type: Optional[FunctionType]
    __dd_wrappers__ = None  # type: Optional[Dict[Any, Any]]

    def __call__(self, *args, **kwargs):
        pass


Wrapper = Callable[[FunctionType, Tuple[Any], Dict[str, Any]], Any]


def _compare_exc(label, lineno):
    """Compat helper for comparing exceptions."""
    return (
        Instr("COMPARE_OP", Compare.EXC_MATCH, lineno=lineno)
        if PY < (3, 9)
        else Instr("JUMP_IF_NOT_EXC_MATCH", label, lineno=lineno)
    )


def _jump_if_false(label, lineno):
    """Compat helper for jumping if false after comparing exceptions."""
    return Instr("POP_JUMP_IF_FALSE", label, lineno=lineno) if PY < (3, 9) else Instr("NOP", lineno=lineno)


def _end_finally(lineno):
    """Compat helper for ending finally blocks."""
    if PY < (3, 9):
        return Instr("END_FINALLY", lineno=lineno)
    elif PY < (3, 10):
        return Instr("RERAISE", lineno=lineno)
    return Instr("RERAISE", 0, lineno=lineno)


# -----------------------------------------------------------------------------
# Generator Wrapping
# -----------------------------------------------------------------------------
# DEV: This is roughly equivalent to
#
# __ddgen = wrapper(wrapped, args, kwargs)
# __ddgensend = __ddgen.send
# try:
#     value = next(__ddgen)
# except StopIteration:
#     return
# while True:
#     try:
#         tosend = yield value
#     except GeneratorExit:
#         return __ddgen.close()
#     except:
#         value = __ddgen.throw(*sys.exc_info())
#     else:
#         try:
#             value = __ddgensend(tosend)
#         except StopIteration:
#             return
# -----------------------------------------------------------------------------
def _wrap_generator(instrs, code, lineno):
    stopiter = Label()
    loop = Label()
    genexit = Label()
    exc = Label()
    propagate = Label()
    _yield = Label()

    instrs[-1:] = [
        Instr("DUP_TOP", lineno=lineno),
        Instr("STORE_FAST", "__ddgen", lineno=lineno),
        Instr("LOAD_ATTR", "send", lineno=lineno),
        Instr("STORE_FAST", "__ddgensend", lineno=lineno),
        Instr("LOAD_CONST", next, lineno=lineno),
        Instr("LOAD_FAST", "__ddgen", lineno=lineno),
        loop,
        Instr("SETUP_EXCEPT" if PY < (3, 8) else "SETUP_FINALLY", stopiter, lineno=lineno),
        Instr("CALL_FUNCTION", 1, lineno=lineno),
        Instr("POP_BLOCK", lineno=lineno),
        _yield,
        Instr("SETUP_EXCEPT" if PY < (3, 8) else "SETUP_FINALLY", genexit, lineno=lineno),
        Instr("YIELD_VALUE", lineno=lineno),
        Instr("POP_BLOCK", lineno=lineno),
        Instr("LOAD_FAST", "__ddgensend", lineno=lineno),
        Instr("ROT_TWO", lineno=lineno),
        Instr("JUMP_ABSOLUTE", loop, lineno=lineno),
        stopiter,  # except StpIteration:
        Instr("DUP_TOP", lineno=lineno),
        Instr("LOAD_CONST", StopIteration, lineno=lineno),
        _compare_exc(propagate, lineno),
        _jump_if_false(propagate, lineno),
        Instr("POP_TOP", lineno=lineno),
        Instr("POP_TOP", lineno=lineno),
        Instr("POP_TOP", lineno=lineno),
        Instr("LOAD_CONST", None, lineno=lineno),
        Instr("RETURN_VALUE", lineno=lineno),
        propagate,
        _end_finally(lineno),
        Instr("LOAD_CONST", None, lineno=lineno),
        Instr("RETURN_VALUE", lineno=lineno),
        genexit,  # except GeneratorExit:
        Instr("DUP_TOP", lineno=lineno),
        Instr("LOAD_CONST", GeneratorExit, lineno=lineno),
        _compare_exc(exc, lineno),
        _jump_if_false(exc, lineno),
        Instr("POP_TOP", lineno=lineno),
        Instr("POP_TOP", lineno=lineno),
        Instr("POP_TOP", lineno=lineno),
        Instr("POP_TOP", lineno=lineno),
        Instr("LOAD_FAST", "__ddgen", lineno=lineno),
        Instr("LOAD_ATTR", "close", lineno=lineno),
        Instr("CALL_FUNCTION", 0, lineno=lineno),
        Instr("RETURN_VALUE", lineno=lineno),
        exc,  # except:
        Instr("POP_TOP", lineno=lineno),
        Instr("POP_TOP", lineno=lineno),
        Instr("POP_TOP", lineno=lineno),
        Instr("POP_TOP", lineno=lineno),
        Instr("LOAD_FAST", "__ddgen", lineno=lineno),
        Instr("LOAD_ATTR", "throw", lineno=lineno),
        Instr("LOAD_CONST", sys.exc_info, lineno=lineno),
        Instr("CALL_FUNCTION", 0, lineno=lineno),
        Instr("CALL_FUNCTION_VAR" if PY < (3, 6) else "CALL_FUNCTION_EX", 0, lineno=lineno),
        # DEV: We cannot use ROT_FOUR because it was removed in 3.5 and added
        # back in 3.8
        Instr("STORE_FAST", "__value", lineno=lineno),
        Instr("POP_EXCEPT" if PY >= (3,) else "NOP", lineno=lineno),
        Instr("LOAD_FAST", "__value", lineno=lineno),
        Instr("JUMP_ABSOLUTE", _yield, lineno=lineno),
    ]


def _wrap_generator_py311(instrs, code, lineno):
    # With the latest changes to the Python opcode we don't seem to need any
    # special handling for generators
    pass


wrap_generator = _wrap_generator_py311 if PY >= (3, 11) else _wrap_generator


# -----------------------------------------------------------------------------
# Coroutine and Async Generator Wrapping
# -----------------------------------------------------------------------------
# DEV: The wrapping of async generators is roughly equivalent to
#
# __ddgen = wrapper(wrapped, args, kwargs)
# __ddgensend = __ddgen.asend
# try:
#     value = await __ddgen.__anext__()
# except StopAsyncIteration:
#     return
# while True:
#     try:
#         tosend = yield value
#     except GeneratorExit:
#         await __ddgen.aclose()
#     except:
#         value = await __ddgen.athrow(*sys.exc_info())
#     else:
#         try:
#             value = await __ddgensend(tosend)
#         except StopAsyncIteration:
#             return
# -----------------------------------------------------------------------------
def _wrap_special_function_py3(instrs, code, lineno):
    if CompilerFlags.COROUTINE & code.co_flags:
        # DEV: This is just
        # >>> return await wrapper(wrapped, args, kwargs)
        instrs[-1:-1] = [
            Instr("GET_AWAITABLE", lineno=lineno),
            Instr("LOAD_CONST", None, lineno=lineno),
            Instr("YIELD_FROM", lineno=lineno),
        ]
    elif CompilerFlags.ASYNC_GENERATOR & code.co_flags:
        stopiter = Label()
        loop = Label()
        genexit = Label()
        exc = Label()
        propagate = Label()
        _yield = Label()

        instrs[-1:] = [
            Instr("DUP_TOP", lineno=lineno),
            Instr("STORE_FAST", "__ddgen", lineno=lineno),
            Instr("LOAD_ATTR", "asend", lineno=lineno),
            Instr("STORE_FAST", "__ddgensend", lineno=lineno),
            Instr("LOAD_FAST", "__ddgen", lineno=lineno),
            Instr("LOAD_ATTR", "__anext__", lineno=lineno),
            Instr("CALL_FUNCTION", 0, lineno=lineno),
            loop,
            Instr("GET_AWAITABLE", lineno=lineno),
            Instr("LOAD_CONST", None, lineno=lineno),
            Instr("SETUP_EXCEPT" if PY < (3, 8) else "SETUP_FINALLY", stopiter, lineno=lineno),
            Instr("YIELD_FROM", lineno=lineno),
            Instr("POP_BLOCK", lineno=lineno),
            _yield,
            Instr("SETUP_EXCEPT" if PY < (3, 8) else "SETUP_FINALLY", genexit, lineno=lineno),
            Instr("YIELD_VALUE", lineno=lineno),
            Instr("POP_BLOCK", lineno=lineno),
            Instr("LOAD_FAST", "__ddgensend", lineno=lineno),
            Instr("ROT_TWO", lineno=lineno),
            Instr("CALL_FUNCTION", 1, lineno=lineno),
            Instr("JUMP_ABSOLUTE", loop, lineno=lineno),
            stopiter,  # except StopAsyncIteration:
            Instr("DUP_TOP", lineno=lineno),
            Instr("LOAD_CONST", StopAsyncIteration, lineno=lineno),
            _compare_exc(propagate, lineno),
            _jump_if_false(propagate, lineno),
            Instr("POP_TOP", lineno=lineno),
            Instr("POP_TOP", lineno=lineno),
            Instr("POP_TOP", lineno=lineno),
            Instr("LOAD_CONST", None, lineno=lineno),
            Instr("RETURN_VALUE", lineno=lineno),
            propagate,  # finally:
            _end_finally(lineno),
            Instr("LOAD_CONST", None, lineno=lineno),
            Instr("RETURN_VALUE", lineno=lineno),
            genexit,  # except GeneratorExit:
            Instr("DUP_TOP", lineno=lineno),
            Instr("LOAD_CONST", GeneratorExit, lineno=lineno),
            _compare_exc(exc, lineno),
            _jump_if_false(exc, lineno),
            Instr("POP_TOP", lineno=lineno),
            Instr("POP_TOP", lineno=lineno),
            Instr("POP_TOP", lineno=lineno),
            Instr("POP_TOP", lineno=lineno),
            Instr("LOAD_FAST", "__ddgen", lineno=lineno),
            Instr("LOAD_ATTR", "aclose", lineno=lineno),
            Instr("CALL_FUNCTION", 0, lineno=lineno),
            Instr("GET_AWAITABLE", lineno=lineno),
            Instr("LOAD_CONST", None, lineno=lineno),
            Instr("YIELD_FROM", lineno=lineno),
            Instr("POP_EXCEPT", lineno=lineno),
            Instr("RETURN_VALUE", lineno=lineno),
            exc,  # except:
            Instr("POP_TOP", lineno=lineno),
            Instr("POP_TOP", lineno=lineno),
            Instr("POP_TOP", lineno=lineno),
            Instr("POP_TOP", lineno=lineno),
            Instr("LOAD_FAST", "__ddgen", lineno=lineno),
            Instr("LOAD_ATTR", "athrow", lineno=lineno),
            Instr("LOAD_CONST", sys.exc_info, lineno=lineno),
            Instr("CALL_FUNCTION", 0, lineno=lineno),
            Instr("CALL_FUNCTION_EX", 0, lineno=lineno),
            Instr("GET_AWAITABLE", lineno=lineno),
            Instr("LOAD_CONST", None, lineno=lineno),
            Instr("YIELD_FROM", lineno=lineno),
            # DEV: We cannot use ROT_FOUR because it was removed in 3.5 and added
            # back in 3.8
            Instr("STORE_FAST", "__value", lineno=lineno),
            Instr("POP_EXCEPT", lineno=lineno),
            Instr("LOAD_FAST", "__value", lineno=lineno),
            Instr("JUMP_ABSOLUTE", _yield, lineno=lineno),
        ]


def _wrap_special_function_py311(instrs, code, lineno):
    # With the latest changes to the Python opcode we don't seem to need any
    # special handling for generators
    pass


wrap_special_function = _wrap_special_function_py3 if sys.version_info < (3, 11) else _wrap_special_function_py311


def wrap_bytecode(wrapper, wrapped):
    # type: (Wrapper, FunctionType) -> Bytecode
    """Wrap a function with a wrapper function.

    The wrapper function expects the wrapped function as the first argument,
    followed by the tuple of arguments and the dictionary of keyword arguments.
    The nature of the wrapped function is also honored, meaning that a generator
    function will return a generator function, and a coroutine function will
    return a coroutine function, and so on. The signature is also preserved to
    avoid breaking, e.g., usages of the ``inspect`` module.
    """

    code = wrapped.__code__
    lineno = code.co_firstlineno + (sys.version_info >= (3, 11))
    varargs = bool(code.co_flags & CompilerFlags.VARARGS)
    varkwargs = bool(code.co_flags & CompilerFlags.VARKEYWORDS)
    nargs = code.co_argcount
    argnames = code.co_varnames[:nargs]
    try:
        kwonlyargs = code.co_kwonlyargcount
    except AttributeError:
        kwonlyargs = 0
    kwonlyargnames = code.co_varnames[nargs : nargs + kwonlyargs]
    varargsname = code.co_varnames[nargs + kwonlyargs] if varargs else None
    varkwargsname = code.co_varnames[nargs + kwonlyargs + varargs] if varkwargs else None

    # Push the wrapper function that is to be called and the wrapped function to
    # be passed as first argument.
    instrs = [
        Instr("LOAD_CONST", wrapper, lineno=lineno),
        Instr("LOAD_CONST", wrapped, lineno=lineno),
    ]
    if sys.version_info >= (3, 11):
        instrs[0:0] = [
            Instr("RESUME", 0, lineno=lineno - 1),
            Instr("PUSH_NULL", lineno=lineno),
        ]

    # Build the tuple of all the positional arguments
    if nargs:
        instrs.extend([Instr("LOAD_FAST", argname, lineno=lineno) for argname in argnames])
        instrs.append(Instr("BUILD_TUPLE", nargs, lineno=lineno))
        if varargs:
            instrs.extend(
                [
                    Instr("LOAD_FAST", varargsname, lineno=lineno),
                    Instr("INPLACE_ADD", lineno=lineno)
                    if sys.version_info < (3, 11)
                    else Instr("BINARY_OP", b.BinaryOp.ADD, lineno=lineno),
                ]
            )
    elif varargs:
        instrs.append(Instr("LOAD_FAST", varargsname, lineno=lineno))
    else:
        instrs.append(Instr("BUILD_TUPLE", 0, lineno=lineno))

    # Prepare the keyword arguments
    if kwonlyargs:
        for arg in kwonlyargnames:
            instrs.extend(
                [
                    Instr("LOAD_CONST", arg, lineno=lineno),
                    Instr("LOAD_FAST", arg, lineno=lineno),
                ]
            )
        instrs.append(Instr("BUILD_MAP", kwonlyargs, lineno=lineno))
        if varkwargs:
            if sys.version_info < (3, 11):
                instrs.extend(
                    [
                        Instr("DUP_TOP", lineno=lineno),
                        Instr("LOAD_ATTR", "update", lineno=lineno),
                        Instr("LOAD_FAST", varkwargsname, lineno=lineno),
                        Instr("CALL_FUNCTION", 1, lineno=lineno),
                        Instr("POP_TOP", lineno=lineno),
                    ]
                )
            else:
                instrs.extend(
                    [
                        Instr("COPY", 1, lineno=lineno),
                        Instr("LOAD_METHOD", "update", lineno=lineno),
                        Instr("LOAD_FAST", varkwargsname, lineno=lineno),
                        Instr("PRECALL", 1, lineno=lineno),
                        Instr("CALL", 1, lineno=lineno),
                        Instr("POP_TOP", lineno=lineno),
                    ]
                )

    elif varkwargs:
        instrs.append(Instr("LOAD_FAST", varkwargsname, lineno=lineno))

    else:
        instrs.append(Instr("BUILD_MAP", 0, lineno=lineno))

    # Call the wrapper function with the wrapped function, the positional and
    # keyword arguments, and return the result.
    if sys.version_info < (3, 11):
        instrs.extend(
            [
                Instr("CALL_FUNCTION", 3, lineno=lineno),
                Instr("RETURN_VALUE", lineno=lineno),
            ]
        )
    else:
        instrs.extend(
            [
                Instr("PRECALL", 3, lineno=lineno),
                Instr("CALL", 3, lineno=lineno),
                Instr("RETURN_VALUE", lineno=lineno),
            ]
        )

    # If the function has special flags set, like the generator, async generator
    # or coroutine, inject unraveling code before the return opcode.
    if CompilerFlags.GENERATOR & code.co_flags and not (CompilerFlags.COROUTINE & code.co_flags):
        wrap_generator(instrs, code, lineno)
    elif PY3:
        wrap_special_function(instrs, code, lineno)

    return Bytecode(instrs)


def wrap(f, wrapper):
    # type: (FunctionType, Wrapper) -> WrappedFunction
    """Wrap a function with a wrapper.

    The wrapper expects the function as first argument, followed by the tuple
    of positional arguments and the dict of keyword arguments.

    Note that this changes the behavior of the original function with the
    wrapper function, instead of creating a new function object.
    """
    wrapped = FunctionType(
        f.__code__,
        f.__globals__,
        "<wrapped>",
        f.__defaults__,
        f.__closure__,
    )
    try:
        wf = cast(WrappedFunction, f)
        cast(WrappedFunction, wrapped).__dd_wrapped__ = cast(FunctionType, wf.__dd_wrapped__)
    except AttributeError:
        pass

    if PY3:
        wrapped.__kwdefaults__ = f.__kwdefaults__

    code = wrap_bytecode(wrapper, wrapped)
    code.freevars = f.__code__.co_freevars
    code.name = f.__code__.co_name
    code.filename = f.__code__.co_filename
    code.flags = f.__code__.co_flags
    code.argcount = f.__code__.co_argcount
    try:
        code.posonlyargcount = f.__code__.co_posonlyargcount
    except AttributeError:
        pass

    nargs = code.argcount
    try:
        code.kwonlyargcount = f.__code__.co_kwonlyargcount
        nargs += code.kwonlyargcount
    except AttributeError:
        pass
    nargs += bool(code.flags & CompilerFlags.VARARGS) + bool(code.flags & CompilerFlags.VARKEYWORDS)
    code.argnames = f.__code__.co_varnames[:nargs]

    f.__code__ = code.to_code()

    # DEV: Multiple wrapping is implemented as a singly-linked list via the
    # __dd_wrapped__ attribute.
    wf = cast(WrappedFunction, f)
    wf.__dd_wrapped__ = wrapped

    return wf


def unwrap(wf, wrapper):
    # type: (WrappedFunction, Wrapper) -> FunctionType
    """Unwrap a wrapped function.

    This is the reverse of :func:`wrap`. In case of multiple wrapping layers,
    this will unwrap the one that uses ``wrapper``. If the function was not
    wrapped with ``wrapper``, it will return the first argument.
    """
    # DEV: Multiple wrapping layers are singly-linked via __dd_wrapped__. When
    # we find the layer that needs to be removed we also have to ensure that we
    # update the link at the deletion site if there is a non-empty tail.
    try:
        inner = cast(FunctionType, wf.__dd_wrapped__)

        # Sanity check
        assert inner.__name__ == "<wrapped>", "Wrapper has wrapped function"

        if wrapper not in cast(FunctionType, wf).__code__.co_consts:
            # This is not the correct wrapping layer. Try with the next one.
            inner_wf = cast(WrappedFunction, inner)
            return unwrap(inner_wf, wrapper)

        # Remove the current wrapping layer by moving the next one over the
        # current one.
        f = cast(FunctionType, wf)
        f.__code__ = inner.__code__
        try:
            # Update the link to the next layer.
            inner_wf = cast(WrappedFunction, inner)
            wf.__dd_wrapped__ = inner_wf.__dd_wrapped__  # type: ignore[assignment]
        except AttributeError:
            # No more wrapping layers. Restore the original function by removing
            # this extra attribute.
            del wf.__dd_wrapped__

        return f

    except AttributeError:
        # The function is not wrapped so we return it as is.
        return cast(FunctionType, wf)



import pytest

from tests.utils import snapshot

from .test_integration import AGENT_VERSION


pytestmark = pytest.mark.skipif(AGENT_VERSION != "testagent", reason="Tests only compatible with a testagent")


@snapshot()
def test_context_multiprocess(run_python_code_in_subprocess):
    # Testing example from our docs:
    # https://ddtrace.readthedocs.io/en/stable/advanced_usage.html#tracing-across-processes
    code = """
from multiprocessing import Process
import time

from ddtrace import tracer


def _target(ctx):
    tracer.context_provider.activate(ctx)
    with tracer.trace("proc"):
        time.sleep(0.1)
    tracer.shutdown()


def main():
    with tracer.trace("work"):
        proc = Process(target=_target, args=(tracer.current_trace_context(), ))
        proc.start()
        time.sleep(0.25)
        proc.join()


if __name__ == "__main__":
    main()
    """

    stdout, stderr, status, _ = run_python_code_in_subprocess(code=code)
    assert status == 0, (stdout, stderr)
    assert stdout == b"", stderr
    assert stderr == b"", stdout



#!/usr/bin/env python3
import sys

from ddtrace.internal.logger import get_logger
from ddtrace.settings import _config as config


def _is_python_version_supported():
    # IAST supports Python versions 3.6 to 3.10
    return (3, 6, 0) <= sys.version_info < (3, 11, 0)


def _is_iast_enabled():
    if not config._iast_enabled:
        return False

    if not _is_python_version_supported():
        log = get_logger(__name__)
        log.info("IAST is not compatible with the current Python version")
        return False

    return True



import inspect
import os

import boto.connection

from ddtrace import config
from ddtrace.constants import ANALYTICS_SAMPLE_RATE_KEY
from ddtrace.constants import SPAN_KIND
from ddtrace.constants import SPAN_MEASURED_KEY
from ddtrace.ext import SpanKind
from ddtrace.ext import SpanTypes
from ddtrace.ext import aws
from ddtrace.ext import http
from ddtrace.internal.constants import COMPONENT
from ddtrace.internal.utils.wrappers import unwrap
from ddtrace.pin import Pin
from ddtrace.vendor import debtcollector
from ddtrace.vendor import wrapt

from ...internal.utils import get_argument_value
from ...internal.utils.formats import asbool


# Original boto client class
_Boto_client = boto.connection.AWSQueryConnection

AWS_QUERY_ARGS_NAME = ("operation_name", "params", "path", "verb")
AWS_AUTH_ARGS_NAME = (
    "method",
    "path",
    "headers",
    "data",
    "host",
    "auth_path",
    "sender",
)
AWS_QUERY_TRACED_ARGS = {"operation_name", "params", "path"}
AWS_AUTH_TRACED_ARGS = {"path", "data", "host"}


if os.getenv("DD_AWS_TAG_ALL_PARAMS") is not None:
    debtcollector.deprecate(
        "Using environment variable 'DD_AWS_TAG_ALL_PARAMS' is deprecated",
        message="The boto integration no longer includes all API parameters by default.",
        removal_version="2.0.0",
    )

config._add(
    "boto",
    {
        "tag_no_params": asbool(os.getenv("DD_AWS_TAG_NO_PARAMS", default=False)),
        "tag_all_params": asbool(os.getenv("DD_AWS_TAG_ALL_PARAMS", default=False)),
    },
)


def patch():
    if getattr(boto.connection, "_datadog_patch", False):
        return
    setattr(boto.connection, "_datadog_patch", True)

    # AWSQueryConnection and AWSAuthConnection are two different classes called by
    # different services for connection.
    # For example EC2 uses AWSQueryConnection and S3 uses AWSAuthConnection
    wrapt.wrap_function_wrapper("boto.connection", "AWSQueryConnection.make_request", patched_query_request)
    wrapt.wrap_function_wrapper("boto.connection", "AWSAuthConnection.make_request", patched_auth_request)
    Pin(service="aws").onto(boto.connection.AWSQueryConnection)
    Pin(service="aws").onto(boto.connection.AWSAuthConnection)


def unpatch():
    if getattr(boto.connection, "_datadog_patch", False):
        setattr(boto.connection, "_datadog_patch", False)
        unwrap(boto.connection.AWSQueryConnection, "make_request")
        unwrap(boto.connection.AWSAuthConnection, "make_request")


# ec2, sqs, kinesis
def patched_query_request(original_func, instance, args, kwargs):

    pin = Pin.get_from(instance)
    if not pin or not pin.enabled():
        return original_func(*args, **kwargs)

    endpoint_name = getattr(instance, "host").split(".")[0]

    with pin.tracer.trace(
        "{}.command".format(endpoint_name),
        service="{}.{}".format(pin.service, endpoint_name),
        span_type=SpanTypes.HTTP,
    ) as span:
        span.set_tag_str(COMPONENT, config.boto.integration_name)

        # set span.kind to the type of request being performed
        span.set_tag_str(SPAN_KIND, SpanKind.CLIENT)

        span.set_tag(SPAN_MEASURED_KEY)

        operation_name = None
        if args:
            operation_name = get_argument_value(args, kwargs, 0, "action")
            params = get_argument_value(args, kwargs, 1, "params")

            span.resource = "%s.%s" % (endpoint_name, operation_name.lower())

            if params and not config.boto["tag_no_params"]:
                aws._add_api_param_span_tags(span, endpoint_name, params)
        else:
            span.resource = endpoint_name

        if not config.boto["tag_no_params"] and config.boto["tag_all_params"]:
            aws.add_span_arg_tags(span, endpoint_name, args, AWS_QUERY_ARGS_NAME, AWS_QUERY_TRACED_ARGS)

        # Obtaining region name
        region_name = _get_instance_region_name(instance)

        meta = {
            aws.AGENT: "boto",
            aws.OPERATION: operation_name,
        }
        if region_name:
            meta[aws.REGION] = region_name

        span.set_tags(meta)

        # Original func returns a boto.connection.HTTPResponse object
        result = original_func(*args, **kwargs)
        span.set_tag(http.STATUS_CODE, getattr(result, "status"))
        span.set_tag_str(http.METHOD, getattr(result, "_method"))

        # set analytics sample rate
        span.set_tag(ANALYTICS_SAMPLE_RATE_KEY, config.boto.get_analytics_sample_rate())

        return result


# s3, lambda
def patched_auth_request(original_func, instance, args, kwargs):

    # Catching the name of the operation that called make_request()
    operation_name = None

    # Go up the stack until we get the first non-ddtrace module
    # DEV: For `lambda.list_functions()` this should be:
    #        - ddtrace.contrib.boto.patch
    #        - ddtrace.vendor.wrapt.wrappers
    #        - boto.awslambda.layer1 (make_request)
    #        - boto.awslambda.layer1 (list_functions)
    # But can vary depending on Python versions; that's why we use an heuristic
    frame = inspect.currentframe().f_back
    operation_name = None
    while frame:
        if frame.f_code.co_name == "make_request":
            operation_name = frame.f_back.f_code.co_name
            break
        frame = frame.f_back

    pin = Pin.get_from(instance)
    if not pin or not pin.enabled():
        return original_func(*args, **kwargs)

    endpoint_name = getattr(instance, "host").split(".")[0]

    with pin.tracer.trace(
        "{}.command".format(endpoint_name),
        service="{}.{}".format(pin.service, endpoint_name),
        span_type=SpanTypes.HTTP,
    ) as span:
        span.set_tag(SPAN_MEASURED_KEY)
        if args:
            http_method = get_argument_value(args, kwargs, 0, "method")
            span.resource = "%s.%s" % (endpoint_name, http_method.lower())
        else:
            span.resource = endpoint_name

        if not config.boto["tag_no_params"] and config.boto["tag_all_params"]:
            aws.add_span_arg_tags(span, endpoint_name, args, AWS_AUTH_ARGS_NAME, AWS_AUTH_TRACED_ARGS)

        # Obtaining region name
        region_name = _get_instance_region_name(instance)

        meta = {
            aws.AGENT: "boto",
            aws.OPERATION: operation_name,
        }
        if region_name:
            meta[aws.REGION] = region_name

        span.set_tags(meta)

        # Original func returns a boto.connection.HTTPResponse object
        result = original_func(*args, **kwargs)
        span.set_tag(http.STATUS_CODE, getattr(result, "status"))
        span.set_tag_str(http.METHOD, getattr(result, "_method"))

        # set analytics sample rate
        span.set_tag(ANALYTICS_SAMPLE_RATE_KEY, config.boto.get_analytics_sample_rate())

        span.set_tag_str(COMPONENT, config.boto.integration_name)

        # set span.kind to the type of request being performed
        span.set_tag_str(SPAN_KIND, SpanKind.CLIENT)

        return result


def _get_instance_region_name(instance):
    region = getattr(instance, "region", None)

    if not region:
        return None
    if isinstance(region, str):
        return region.split(":")[1]
    else:
        return region.name



import time

from datadog_lambda.wrapper import datadog_lambda_wrapper

from ddtrace import tracer


@datadog_lambda_wrapper
def manually_wrapped_handler(event, context):
    @tracer.wrap("result-trace")
    def get_result():
        return {"success": True}

    return get_result()


def handler(event, context):
    return {"success": True}


def timeout_handler(event, context):
    @tracer.wrap("timeout")
    def timeout():
        time.sleep(3)

    timeout()

    return {"success": False}


def finishing_spans_early_handler(event, context):
    root_span = tracer.current_root_span()
    root_span.finish()

    time.sleep(0.5)

    return {"success": False}


def datadog(_handler):
    @tracer.wrap("aws.lambda")
    def wrapped(*args):
        return _handler(*args)

    return wrapped



import pytest

from ddtrace.appsec.iast import oce
from ddtrace.appsec.iast.taint_sinks.weak_cipher import patch as weak_cipher_patch
from ddtrace.appsec.iast.taint_sinks.weak_cipher import unpatch_iast as weak_cipher_unpatch
from ddtrace.appsec.iast.taint_sinks.weak_hash import patch as weak_hash_patch
from ddtrace.appsec.iast.taint_sinks.weak_hash import unpatch_iast as weak_hash_unpatch
from tests.utils import override_env


def iast_span(tracer, env):
    with override_env(env):
        with tracer.trace("test") as span:
            weak_hash_patch()
            weak_cipher_patch()
            oce.acquire_request()
            yield span
            oce.release_request()
            weak_hash_unpatch()
            weak_cipher_unpatch()


@pytest.fixture
def iast_span_defaults(tracer):
    for t in iast_span(tracer, dict(DD_IAST_ENABLED="true")):
        yield t


@pytest.fixture
def iast_span_des_rc2_configured(tracer):
    for t in iast_span(tracer, dict(DD_IAST_ENABLED="true", DD_IAST_WEAK_CIPHER_ALGORITHMS="DES, RC2")):
        yield t


@pytest.fixture
def iast_span_rc4_configured(tracer):
    for t in iast_span(tracer, dict(DD_IAST_ENABLED="true", DD_IAST_WEAK_CIPHER_ALGORITHMS="RC4")):
        yield t


@pytest.fixture
def iast_span_blowfish_configured(tracer):
    for t in iast_span(tracer, dict(DD_IAST_ENABLED="true", DD_IAST_WEAK_CIPHER_ALGORITHMS="BLOWFISH, RC2")):
        yield t


@pytest.fixture
def iast_span_md5_and_sha1_configured(tracer):
    for t in iast_span(tracer, dict(DD_IAST_ENABLED="true", DD_IAST_WEAK_HASH_ALGORITHMS="MD5, SHA1")):
        yield t


@pytest.fixture
def iast_span_only_md4(tracer):
    for t in iast_span(tracer, dict(DD_IAST_ENABLED="true", DD_IAST_WEAK_HASH_ALGORITHMS="MD4")):
        yield t


@pytest.fixture
def iast_span_only_md5(tracer):
    for t in iast_span(tracer, dict(DD_IAST_ENABLED="true", DD_IAST_WEAK_HASH_ALGORITHMS="MD5")):
        yield t


@pytest.fixture
def iast_span_only_sha1(tracer):
    for t in iast_span(tracer, dict(DD_IAST_ENABLED="true", DD_IAST_WEAK_HASH_ALGORITHMS="SHA1")):
        yield t



import os
import subprocess

import pytest


@pytest.mark.parametrize(
    "env_var_name,env_var_value,expected_obfuscation_config,expected_global_query_string_obfuscation_disabled,"
    "expected_http_tag_query_string",
    [
        ("DD_TRACE_OBFUSCATION_QUERY_STRING_PATTERN", "", None, True, True),
        (
            "DD_TRACE_OBFUSCATION_QUERY_STRING_PATTERN",
            "(?i)(?:p(?:ass)?w(?:or))",
            "re.compile('(?i)(?:p(?:ass)?w(?:or))'.encode('ascii'))",
            False,
            True,
        ),
        (
            "DD_WRONG_ENV_NAME",
            "(?i)(?:p(?:ass)?w(?:or))",
            "re.compile(DD_TRACE_OBFUSCATION_QUERY_STRING_PATTERN_DEFAULT.encode('ascii'))",
            False,
            True,
        ),
        (
            "DD_TRACE_OBFUSCATION_QUERY_STRING_PATTERN",
            "(",
            None,
            False,
            False,
        ),
    ],
)
def test_obfuscation_querystring_pattern_env_var(
    env_var_name,
    env_var_value,
    expected_obfuscation_config,
    expected_global_query_string_obfuscation_disabled,
    expected_http_tag_query_string,
):
    """
    Test that obfuscation config is properly configured from env vars
    """
    env = os.environ.copy()
    env[env_var_name] = env_var_value
    out = subprocess.check_output(
        [
            "python",
            "-c",
            (
                """import re;from ddtrace import config;
from ddtrace.settings.config import DD_TRACE_OBFUSCATION_QUERY_STRING_PATTERN_DEFAULT;
assert config._obfuscation_query_string_pattern == %s;
assert config.global_query_string_obfuscation_disabled == %s;
assert config.http_tag_query_string == %s
"""
                % (
                    expected_obfuscation_config,
                    expected_global_query_string_obfuscation_disabled,
                    expected_http_tag_query_string,
                )
            ),
        ],
        env=env,
    )
    assert b"AssertionError" not in out


@pytest.mark.parametrize(
    "server_tag_query_string,client_tag_query_string,"
    "django_expected_http_tag_query_string,requests_expected_http_tag_query_string",
    [
        ("True", "true", True, True),
        (None, None, True, True),
        ("", None, True, True),
        (None, "", True, True),
        ("invalid", None, True, True),
        (None, "invalid", True, True),
        ("1", None, True, True),
        ("True", None, True, True),
        ("0", None, False, True),
        ("False", None, False, True),
        (None, "0", True, False),
        (None, "FALSE", True, False),
    ],
)
def test_tag_querystring_env_var(
    server_tag_query_string,
    client_tag_query_string,
    django_expected_http_tag_query_string,
    requests_expected_http_tag_query_string,
):
    """
    Test that query string tagging config is properly configured from env vars
    """
    env = os.environ.copy()
    if server_tag_query_string is not None:
        env["DD_HTTP_SERVER_TAG_QUERY_STRING"] = server_tag_query_string
    if client_tag_query_string is not None:
        env["DD_HTTP_CLIENT_TAG_QUERY_STRING"] = client_tag_query_string
    out = subprocess.check_output(
        [
            "python",
            "-c",
            (
                """
import os;
from ddtrace import config;
config._add(
    "requests",
    {
        "default_http_tag_query_string": os.getenv("DD_HTTP_CLIENT_TAG_QUERY_STRING", "true"),
    },
);
assert config.django.http_tag_query_string == %s;
assert config.requests.http_tag_query_string == %s
"""
                % (
                    django_expected_http_tag_query_string,
                    requests_expected_http_tag_query_string,
                )
            ),
        ],
        env=env,
    )
    assert b"AssertionError" not in out



import base64
import hashlib
import json
import os
import sys
import time

import mock
from mock.mock import ANY
import pytest

from ddtrace.appsec import _asm_request_context
from ddtrace.appsec._constants import PRODUCTS
from ddtrace.appsec._remoteconfiguration import RCAppSecCallBack
from ddtrace.appsec._remoteconfiguration import RCAppSecFeaturesCallBack
from ddtrace.appsec._remoteconfiguration import _appsec_rules_data
from ddtrace.appsec._remoteconfiguration import enable_appsec_rc
from ddtrace.appsec.processor import AppSecSpanProcessor
from ddtrace.appsec.utils import _appsec_rc_capabilities
from ddtrace.appsec.utils import _appsec_rc_features_is_enabled
from ddtrace.constants import APPSEC_ENV
from ddtrace.constants import APPSEC_JSON
from ddtrace.contrib.trace_utils import set_http_meta
from ddtrace.ext import SpanTypes
from ddtrace.internal import _context
from ddtrace.internal.remoteconfig import RemoteConfig
from ddtrace.internal.remoteconfig.client import AgentPayload
from ddtrace.internal.remoteconfig.client import ConfigMetadata
from ddtrace.internal.remoteconfig.client import TargetFile
from tests.appsec.test_processor import Config
from tests.appsec.test_processor import ROOT_DIR
from tests.utils import override_env
from tests.utils import override_global_config


def _stop_remote_config_worker():
    if RemoteConfig._worker:
        RemoteConfig._worker._stop_service()
        RemoteConfig._worker = None


@pytest.fixture
def remote_config_worker(tracer):
    _stop_remote_config_worker()
    try:
        yield
    finally:
        _stop_remote_config_worker()


def _set_and_get_appsec_tags(tracer):
    with tracer.trace("test", span_type=SpanTypes.WEB) as span:
        set_http_meta(
            span,
            {},
            raw_uri="http://example.com/.git",
            status_code="404",
            request_cookies={"cookie1": "im the cookie1"},
        )
    return span.get_tag(APPSEC_JSON)


def test_rc_enabled_by_default(tracer):
    os.environ.pop("DD_REMOTE_CONFIGURATION_ENABLED", None)
    result = _set_and_get_appsec_tags(tracer)
    assert result is None
    assert _appsec_rc_features_is_enabled()


def test_rc_activate_is_active_and_get_processor_tags(tracer, remote_config_worker):
    result = _set_and_get_appsec_tags(tracer)
    assert result is None
    RCAppSecFeaturesCallBack(tracer)(None, {"asm": {"enabled": True}})
    assert "triggers" in _set_and_get_appsec_tags(tracer)


@pytest.mark.parametrize(
    "appsec_enabled, rc_value",
    [
        ("", True),
        ("true", True),
        ("true", False),
    ],
)
def test_rc_activation_states_on(tracer, appsec_enabled, rc_value, remote_config_worker):
    with override_global_config(dict(_appsec_enabled=False)), override_env({APPSEC_ENV: appsec_enabled}):
        if appsec_enabled == "":
            del os.environ[APPSEC_ENV]

        RCAppSecFeaturesCallBack(tracer)(None, {"asm": {"enabled": rc_value}})
        result = _set_and_get_appsec_tags(tracer)
        assert result
        assert "triggers" in result


@pytest.mark.parametrize(
    "appsec_enabled, rc_value",
    [
        ("", False),
        ("false", False),
        ("false", True),
    ],
)
def test_rc_activation_states_off(tracer, appsec_enabled, rc_value, remote_config_worker):
    with override_global_config(dict(_appsec_enabled=True)), override_env({APPSEC_ENV: appsec_enabled}):
        if appsec_enabled == "":
            del os.environ[APPSEC_ENV]

        rc_config = {"asm": {"enabled": True}}
        if rc_value is False:
            rc_config = False

        RCAppSecFeaturesCallBack(tracer)(None, rc_config)
        result = _set_and_get_appsec_tags(tracer)
        assert result is None


@pytest.mark.parametrize(
    "rc_enabled, appsec_enabled, capability",
    [
        ("true", "true", "Afw="),
        ("false", "true", "Afw="),
        ("true", "false", "Afw="),
        ("false", "false", ""),
        ("true", "", "Af4="),
        ("false", "", ""),
    ],
)
def test_rc_capabilities(rc_enabled, appsec_enabled, capability):
    env = {"DD_REMOTE_CONFIGURATION_ENABLED": rc_enabled}
    if appsec_enabled:
        env[APPSEC_ENV] = appsec_enabled
    with override_env(env):
        assert _appsec_rc_capabilities() == capability


def test_rc_activation_validate_products(tracer, remote_config_worker):
    with override_global_config(dict(_appsec_enabled=False, api_version="v0.4")):
        rc_config = {"asm": {"enabled": True}}

        assert not RemoteConfig._worker

        RCAppSecFeaturesCallBack(tracer)(None, rc_config)

        assert RemoteConfig._worker._client._products["ASM_DATA"]


def test_rc_activation_check_asm_features_product_disables_rest_of_products(tracer, remote_config_worker):
    with override_global_config(dict(_appsec_enabled=True, api_version="v0.4")):
        tracer.configure(appsec_enabled=True, api_version="v0.4")
        enable_appsec_rc(tracer)

        assert RemoteConfig._worker._client._products.get(PRODUCTS.ASM_DATA)
        assert RemoteConfig._worker._client._products.get(PRODUCTS.ASM)
        assert RemoteConfig._worker._client._products.get(PRODUCTS.ASM_FEATURES)

        RCAppSecFeaturesCallBack(tracer)(None, False)

        assert RemoteConfig._worker._client._products.get(PRODUCTS.ASM_DATA) is None
        assert RemoteConfig._worker._client._products.get(PRODUCTS.ASM) is None
        assert RemoteConfig._worker._client._products.get(PRODUCTS.ASM_FEATURES)


def test_load_new_configurations_invalid_content(remote_config_worker, tracer):
    with override_global_config(dict(_appsec_enabled=True, api_version="v0.4")):
        tracer.configure(appsec_enabled=True, api_version="v0.4")
        enable_appsec_rc(tracer)
        asm_features_data = b'{"asm":{"enabled":true}}'
        asm_data_data = b'{"data": "data"}'
        payload = AgentPayload(
            target_files=[
                TargetFile(path="mock/ASM_FEATURES", raw=base64.b64encode(asm_features_data)),
                TargetFile(path="mock/ASM_DATA", raw=base64.b64encode(asm_data_data)),
            ]
        )
        client_configs = {
            "mock/ASM_FEATURES": ConfigMetadata(
                id="",
                product_name="ASM_FEATURES",
                sha256_hash=hashlib.sha256(asm_features_data).hexdigest(),
                length=5,
                tuf_version=5,
            ),
            "mock/ASM_DATA": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data).hexdigest(),
                length=5,
                tuf_version=5,
            ),
        }
        with pytest.raises(ValueError):
            RemoteConfig._worker._client._load_new_configurations({}, client_configs, payload=payload)


@mock.patch.object(RCAppSecFeaturesCallBack, "_appsec_1click_activation")
@mock.patch("ddtrace.appsec._remoteconfiguration._appsec_rules_data")
def test_load_new_configurations_dispatch_applied_configs(
    mock_appsec_rules_data, mock_appsec_1click_activation, remote_config_worker, tracer
):
    with override_global_config(dict(_appsec_enabled=True, api_version="v0.4")):
        tracer.configure(appsec_enabled=True, api_version="v0.4")
        enable_appsec_rc(tracer)
        asm_features_data = b'{"asm":{"enabled":true}}'
        asm_data_data = b'{"data": [{"test": "data"}]}'
        payload = AgentPayload(
            target_files=[
                TargetFile(path="mock/ASM_FEATURES", raw=base64.b64encode(asm_features_data)),
                TargetFile(path="mock/ASM_DATA", raw=base64.b64encode(asm_data_data)),
            ]
        )
        client_configs = {
            "mock/ASM_FEATURES": ConfigMetadata(
                id="",
                product_name="ASM_FEATURES",
                sha256_hash=hashlib.sha256(asm_features_data).hexdigest(),
                length=5,
                tuf_version=5,
            ),
            "mock/ASM_DATA": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data).hexdigest(),
                length=5,
                tuf_version=5,
            ),
        }

        RemoteConfig._worker._client._load_new_configurations({}, client_configs, payload=payload)
        mock_appsec_rules_data.assert_called_with(ANY, {"data": [{"test": "data"}]})
        mock_appsec_1click_activation.assert_called_with({"asm": {"enabled": True}})


@mock.patch.object(RCAppSecFeaturesCallBack, "_appsec_1click_activation")
@mock.patch("ddtrace.appsec._remoteconfiguration._appsec_rules_data")
def test_load_new_configurations_empty_config(
    mock_appsec_rules_data, mock_appsec_1click_activation, remote_config_worker, tracer
):
    with override_global_config(dict(_appsec_enabled=True, api_version="v0.4")):
        tracer.configure(appsec_enabled=True, api_version="v0.4")
        enable_appsec_rc(tracer)
        asm_features_data = b'{"asm":{"enabled":true}}'
        asm_data_data = b'{"data": []}'
        payload = AgentPayload(
            target_files=[
                TargetFile(path="mock/ASM_FEATURES", raw=base64.b64encode(asm_features_data)),
                TargetFile(path="mock/ASM_DATA", raw=base64.b64encode(asm_data_data)),
            ]
        )
        client_configs = {
            "mock/ASM_FEATURES": ConfigMetadata(
                id="",
                product_name="ASM_FEATURES",
                sha256_hash=hashlib.sha256(asm_features_data).hexdigest(),
                length=5,
                tuf_version=5,
            ),
            "mock/ASM_DATA": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data).hexdigest(),
                length=5,
                tuf_version=5,
            ),
        }

        RemoteConfig._worker._client._load_new_configurations({}, client_configs, payload=payload)
        mock_appsec_rules_data.assert_called_with(ANY, {"data": []})
        mock_appsec_1click_activation.assert_called_with({"asm": {"enabled": True}})


@mock.patch.object(RCAppSecFeaturesCallBack, "_appsec_1click_activation")
@mock.patch("ddtrace.appsec._remoteconfiguration._appsec_rules_data")
def test_load_new_configurations_remove_config_and_dispatch_applied_configs(
    mock_appsec_rules_data, mock_appsec_1click_activation, remote_config_worker
):
    enable_appsec_rc()
    asm_features_data = b'{"asm":{"enabled":true}}'
    asm_data_data = b'{"data":{}}'
    payload = AgentPayload(
        target_files=[
            TargetFile(path="mock/ASM_FEATURES", raw=base64.b64encode(asm_features_data)),
            TargetFile(path="mock/ASM_DATA", raw=base64.b64encode(asm_data_data)),
        ]
    )
    client_configs = {
        "mock/ASM_FEATURES": ConfigMetadata(
            id="",
            product_name="ASM_FEATURES",
            sha256_hash=hashlib.sha256(asm_features_data).hexdigest(),
            length=5,
            tuf_version=5,
        ),
        "mock/ASM_DATA": ConfigMetadata(
            id="",
            product_name="ASM_DATA",
            sha256_hash=hashlib.sha256(asm_data_data).hexdigest(),
            length=5,
            tuf_version=5,
        ),
    }

    RemoteConfig._worker._client._applied_configs = client_configs
    RemoteConfig._worker._client._remove_previously_applied_configurations({}, {}, [])

    mock_appsec_1click_activation.assert_called_with(False)
    mock_appsec_rules_data.assert_not_called()
    mock_appsec_1click_activation.reset_mock()
    mock_appsec_rules_data.reset_mock()

    # Not called because with this configuration the above condition is True:
    # applied_config = self._applied_configs.get(target)
    # if applied_config == config:
    #     continue
    RemoteConfig._worker._client._load_new_configurations({}, client_configs, payload=payload)
    mock_appsec_rules_data.assert_not_called()
    mock_appsec_1click_activation.assert_not_called()


def test_load_new_configurations_remove_config_and_dispatch_applied_configs_error(remote_config_worker):
    """
    The previous code raises a key error in `self._products[config.product_name]` when appsec features is disabled
    with ASM_FEATURES product and then loops over the config in _load_new_configurations
    """
    enable_appsec_rc()
    asm_features_data = b'{"asm":{"enabled":true}}'
    asm_data_data = b'{"data":{}}'
    payload = AgentPayload(
        target_files=[
            TargetFile(path="mock/ASM_FEATURES", raw=base64.b64encode(asm_features_data)),
            TargetFile(path="mock/ASM_DATA", raw=base64.b64encode(asm_data_data)),
        ]
    )
    client_configs = {
        "mock/ASM_FEATURES": ConfigMetadata(
            id="",
            product_name="ASM_FEATURES",
            sha256_hash=hashlib.sha256(asm_features_data).hexdigest(),
            length=5,
            tuf_version=5,
        ),
        "mock/ASM_DATA": ConfigMetadata(
            id="",
            product_name="ASM_DATA",
            sha256_hash=hashlib.sha256(asm_data_data).hexdigest(),
            length=5,
            tuf_version=5,
        ),
    }

    RemoteConfig._worker._client._applied_configs = client_configs
    RemoteConfig._worker._client._remove_previously_applied_configurations({}, {}, [])

    RemoteConfig._worker._client._load_new_configurations({}, client_configs, payload=payload)


@pytest.mark.skipif(sys.version_info[:2] < (3, 6), reason="Mock return order is different in python <= 3.5")
@mock.patch.object(RCAppSecFeaturesCallBack, "_appsec_1click_activation")
@mock.patch("ddtrace.appsec._remoteconfiguration._appsec_rules_data")
def test_load_multiple_targets_file_same_product(
    mock_appsec_rules_data, mock_appsec_1click_activation, remote_config_worker, tracer
):
    with override_global_config(dict(_appsec_enabled=True, api_version="v0.4")):
        tracer.configure(appsec_enabled=True, api_version="v0.4")
        enable_appsec_rc(tracer)
        asm_features_data = b'{"asm":{"enabled":true}}'
        asm_data_data1 = b'{"data": [{"a":1}]}'
        asm_data_data2 = b'{"data": [{"b":2}]}'
        payload = AgentPayload(
            target_files=[
                TargetFile(path="mock/ASM_FEATURES", raw=base64.b64encode(asm_features_data)),
                TargetFile(path="mock/ASM_DATA/1", raw=base64.b64encode(asm_data_data1)),
                TargetFile(path="mock/ASM_DATA/2", raw=base64.b64encode(asm_data_data2)),
            ]
        )
        client_configs = {
            "mock/ASM_FEATURES": ConfigMetadata(
                id="",
                product_name="ASM_FEATURES",
                sha256_hash=hashlib.sha256(asm_features_data).hexdigest(),
                length=5,
                tuf_version=5,
            ),
            "mock/ASM_DATA/1": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data1).hexdigest(),
                length=5,
                tuf_version=5,
            ),
            "mock/ASM_DATA/2": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data2).hexdigest(),
                length=5,
                tuf_version=5,
            ),
        }

        RemoteConfig._worker._client._load_new_configurations({}, client_configs, payload=payload)
        mock_appsec_rules_data.assert_called_with(ANY, {"data": [{"a": 1}, {"b": 2}]})
        mock_appsec_1click_activation.assert_called_with({"asm": {"enabled": True}})


@mock.patch("ddtrace.appsec._remoteconfiguration._appsec_rules_data")
def test_remove_targets_file_with_previous_configuration(mock_appsec_rules_data, remote_config_worker, tracer):
    with override_global_config(dict(_appsec_enabled=True, api_version="v0.4")):
        RCAppSecCallBack.configs = {}
        tracer.configure(appsec_enabled=True, api_version="v0.4")
        enable_appsec_rc(tracer)
        asm_data_data1 = b'{"data": [{"a":1}]}'
        asm_data_data2 = b'{"data": [{"b":2}]}'
        payload = AgentPayload(
            target_files=[
                TargetFile(path="mock/ASM_DATA/1", raw=base64.b64encode(asm_data_data1)),
                TargetFile(path="mock/ASM_DATA/2", raw=base64.b64encode(asm_data_data2)),
            ]
        )
        applied_configs = {
            "mock/ASM_DATA/1": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data1).hexdigest(),
                length=5,
                tuf_version=5,
            ),
            "mock/ASM_DATA/2": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data2).hexdigest(),
                length=5,
                tuf_version=5,
            ),
        }

        client_configs = {
            "mock/ASM_DATA/1": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data1).hexdigest(),
                length=5,
                tuf_version=5,
            ),
        }

        target_file = {
            "mock/ASM_DATA/2": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data2).hexdigest(),
                length=5,
                tuf_version=5,
            )
        }

        RemoteConfig._worker._client._applied_configs = applied_configs
        RemoteConfig._worker._client._remove_previously_applied_configurations({}, client_configs, target_file)

        RemoteConfig._worker._client._load_new_configurations({}, client_configs, payload=payload)
        mock_appsec_rules_data.assert_not_called()


@pytest.mark.skipif(sys.version_info[:2] < (3, 6), reason="Mock return order is different in python <= 3.5")
@mock.patch.object(RCAppSecFeaturesCallBack, "_appsec_1click_activation")
@mock.patch("ddtrace.appsec._remoteconfiguration._appsec_rules_data")
def test_load_new_config_and_remove_targets_file_same_product(
    mock_appsec_rules_data, mock_appsec_1click_activation, remote_config_worker, tracer
):
    with override_global_config(dict(_appsec_enabled=True, api_version="v0.4")):
        tracer.configure(appsec_enabled=True, api_version="v0.4")
        applied_configs = {}
        enable_appsec_rc(tracer)
        asm_features_data = b'{"asm":{"enabled":true}}'
        asm_data_data1 = b'{"data": [{"a":1}]}'
        asm_data_data2 = b'{"data": [{"b":2}]}'
        payload = AgentPayload(
            target_files=[
                TargetFile(path="mock/ASM_FEATURES", raw=base64.b64encode(asm_features_data)),
                TargetFile(path="mock/ASM_DATA/1", raw=base64.b64encode(asm_data_data1)),
                TargetFile(path="mock/ASM_DATA/2", raw=base64.b64encode(asm_data_data2)),
            ]
        )
        first_config = {
            "mock/ASM_FEATURES": ConfigMetadata(
                id="",
                product_name="ASM_FEATURES",
                sha256_hash=hashlib.sha256(asm_features_data).hexdigest(),
                length=5,
                tuf_version=5,
            ),
            "mock/ASM_DATA/1": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data1).hexdigest(),
                length=5,
                tuf_version=5,
            ),
            "mock/ASM_DATA/2": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data2).hexdigest(),
                length=5,
                tuf_version=5,
            ),
        }

        second_config = {
            "mock/ASM_FEATURES": ConfigMetadata(
                id="",
                product_name="ASM_FEATURES",
                sha256_hash=hashlib.sha256(asm_features_data).hexdigest(),
                length=5,
                tuf_version=5,
            ),
            "mock/ASM_DATA/1": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data1).hexdigest(),
                length=5,
                tuf_version=5,
            ),
        }

        target_file = {
            "mock/ASM_DATA/2": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data2).hexdigest(),
                length=5,
                tuf_version=5,
            )
        }
        RemoteConfig._worker._client._remove_previously_applied_configurations({}, first_config, {})

        RemoteConfig._worker._client._load_new_configurations(applied_configs, first_config, payload=payload)
        RemoteConfig._worker._client._applied_configs = applied_configs

        mock_appsec_rules_data.assert_called_with(ANY, {"data": [{"a": 1}, {"b": 2}]})
        mock_appsec_rules_data.reset_mock()

        RemoteConfig._worker._client._remove_previously_applied_configurations({}, second_config, target_file)

        RemoteConfig._worker._client._load_new_configurations({}, second_config, payload=payload)
        mock_appsec_rules_data.assert_called_with(ANY, {"data": [{"a": 1}]})


@pytest.mark.skipif(sys.version_info[:2] < (3, 6), reason="Mock return order is different in python <= 3.5")
@mock.patch.object(AppSecSpanProcessor, "_update_rules")
def test_fullpath_appsec_rules_data(mock_update_rules, remote_config_worker, tracer):
    with override_global_config(dict(_appsec_enabled=True, api_version="v0.4")):
        tracer.configure(appsec_enabled=True, api_version="v0.4")
        applied_configs = {}
        enable_appsec_rc(tracer)
        asm_features_data = b'{"asm":{"enabled":true}}'
        asm_data_data1 = b'{"exclusions": [{"a":1}]}'
        asm_data_data2 = b'{"exclusions": [{"b":2}]}'
        payload = AgentPayload(
            target_files=[
                TargetFile(path="mock/ASM_FEATURES", raw=base64.b64encode(asm_features_data)),
                TargetFile(path="mock/ASM_DATA/1", raw=base64.b64encode(asm_data_data1)),
                TargetFile(path="mock/ASM_DATA/2", raw=base64.b64encode(asm_data_data2)),
            ]
        )
        first_config = {
            "mock/ASM_FEATURES": ConfigMetadata(
                id="",
                product_name="ASM_FEATURES",
                sha256_hash=hashlib.sha256(asm_features_data).hexdigest(),
                length=5,
                tuf_version=5,
            ),
            "mock/ASM_DATA/1": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data1).hexdigest(),
                length=5,
                tuf_version=5,
            ),
            "mock/ASM_DATA/2": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data2).hexdigest(),
                length=5,
                tuf_version=5,
            ),
        }

        second_config = {
            "mock/ASM_FEATURES": ConfigMetadata(
                id="",
                product_name="ASM_FEATURES",
                sha256_hash=hashlib.sha256(asm_features_data).hexdigest(),
                length=5,
                tuf_version=5,
            ),
            "mock/ASM_DATA/1": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data1).hexdigest(),
                length=5,
                tuf_version=5,
            ),
        }

        target_file = {
            "mock/ASM_DATA/2": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data2).hexdigest(),
                length=5,
                tuf_version=5,
            )
        }
        RemoteConfig._worker._client._remove_previously_applied_configurations({}, first_config, {})

        RemoteConfig._worker._client._load_new_configurations(applied_configs, first_config, payload=payload)
        RemoteConfig._worker._client._applied_configs = applied_configs

        mock_update_rules.assert_called_with({"exclusions": [{"a": 1}, {"b": 2}]})
        mock_update_rules.reset_mock()

        RemoteConfig._worker._client._remove_previously_applied_configurations({}, second_config, target_file)

        RemoteConfig._worker._client._load_new_configurations({}, second_config, payload=payload)
        mock_update_rules.assert_called_with({"exclusions": [{"a": 1}]})


@pytest.mark.skipif(sys.version_info[:2] < (3, 6), reason="Mock return order is different in python <= 3.5")
@mock.patch.object(AppSecSpanProcessor, "_update_rules")
def test_fullpath_appsec_rules_data_empty_data(mock_update_rules, remote_config_worker, tracer):
    with override_global_config(dict(_appsec_enabled=True, api_version="v0.4")):
        tracer.configure(appsec_enabled=True, api_version="v0.4")
        applied_configs = {}
        enable_appsec_rc(tracer)
        asm_data_data1 = b'{"exclusions": [{"a":1}]}'
        asm_data_data2 = b'{"exclusions": []}'
        payload = AgentPayload(
            target_files=[
                TargetFile(path="mock/ASM_DATA/1", raw=base64.b64encode(asm_data_data1)),
                TargetFile(path="mock/ASM_DATA/2", raw=base64.b64encode(asm_data_data2)),
            ]
        )
        first_config = {
            "mock/ASM_DATA/1": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data1).hexdigest(),
                length=5,
                tuf_version=5,
            ),
            "mock/ASM_DATA/2": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data2).hexdigest(),
                length=5,
                tuf_version=5,
            ),
        }

        second_config = {
            "mock/ASM_DATA/1": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data1).hexdigest(),
                length=5,
                tuf_version=5,
            ),
        }

        target_file = {
            "mock/ASM_DATA/2": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data2).hexdigest(),
                length=5,
                tuf_version=5,
            )
        }
        RemoteConfig._worker._client._remove_previously_applied_configurations({}, first_config, {})

        RemoteConfig._worker._client._load_new_configurations(applied_configs, first_config, payload=payload)
        RemoteConfig._worker._client._applied_configs = applied_configs

        mock_update_rules.assert_called_with({"exclusions": [{"a": 1}]})
        mock_update_rules.reset_mock()

        RemoteConfig._worker._client._remove_previously_applied_configurations({}, second_config, target_file)

        RemoteConfig._worker._client._load_new_configurations({}, second_config, payload=payload)
        mock_update_rules.assert_called_with({"exclusions": [{"a": 1}]})


@pytest.mark.skipif(sys.version_info[:2] < (3, 6), reason="Mock return order is different in python <= 3.5")
@mock.patch.object(AppSecSpanProcessor, "_update_rules")
def test_fullpath_appsec_rules_data_add_delete_file(mock_update_rules, remote_config_worker, tracer):
    with override_global_config(dict(_appsec_enabled=True, api_version="v0.4")):
        tracer.configure(appsec_enabled=True, api_version="v0.4")
        applied_configs = {}
        enable_appsec_rc(tracer)
        asm_data_data1 = b'{"exclusions": [{"a":1}]}'
        asm_data_data2 = b'{"exclusions": []}'
        payload = AgentPayload(
            target_files=[
                TargetFile(path="mock/ASM_DATA/1", raw=base64.b64encode(asm_data_data1)),
            ]
        )
        first_config = {
            "mock/ASM_DATA/1": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data1).hexdigest(),
                length=5,
                tuf_version=5,
            ),
        }

        second_payload = AgentPayload(
            target_files=[
                TargetFile(path="mock/ASM_DATA/1", raw=base64.b64encode(asm_data_data2)),
            ]
        )
        second_config = {
            "mock/ASM_DATA/1": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data2).hexdigest(),
                length=5,
                tuf_version=5,
            ),
        }

        target_file = {
            "mock/ASM_DATA/1": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data1).hexdigest(),
                length=5,
                tuf_version=5,
            )
        }
        RemoteConfig._worker._client._remove_previously_applied_configurations({}, first_config, {})

        RemoteConfig._worker._client._load_new_configurations(applied_configs, first_config, payload=payload)
        RemoteConfig._worker._client._applied_configs = applied_configs

        mock_update_rules.assert_called_with({"exclusions": [{"a": 1}]})
        mock_update_rules.reset_mock()

        RemoteConfig._worker._client._remove_previously_applied_configurations({}, second_config, target_file)

        RemoteConfig._worker._client._load_new_configurations({}, second_config, payload=second_payload)
        mock_update_rules.assert_called_with({"exclusions": []})


@pytest.mark.skipif(sys.version_info[:2] < (3, 6), reason="Mock return order is different in python <= 3.5")
@mock.patch.object(RCAppSecFeaturesCallBack, "_appsec_1click_activation")
@mock.patch("ddtrace.appsec._remoteconfiguration._appsec_rules_data")
def test_load_new_empty_config_and_remove_targets_file_same_product(
    mock_appsec_rules_data, remote_config_worker, tracer
):
    with override_global_config(dict(_appsec_enabled=True, api_version="v0.4")):
        tracer.configure(appsec_enabled=True, api_version="v0.4")
        RCAppSecCallBack.configs = {}
        applied_configs = {}
        enable_appsec_rc(tracer)
        asm_features_data = b'{"asm":{"enabled":true}}'
        asm_data_data1 = b'{"data": [{"a":1}]}'
        asm_data_data2 = b'{"data2": [{"b":2}]}'
        asm_data_data_empty = b'{"data2": []}'
        payload = AgentPayload(
            target_files=[
                TargetFile(path="mock/ASM_FEATURES", raw=base64.b64encode(asm_features_data)),
                TargetFile(path="mock/ASM_DATA/1", raw=base64.b64encode(asm_data_data1)),
                TargetFile(path="mock/ASM_DATA/2", raw=base64.b64encode(asm_data_data2)),
            ]
        )
        first_config = {
            "mock/ASM_FEATURES": ConfigMetadata(
                id="",
                product_name="ASM_FEATURES",
                sha256_hash=hashlib.sha256(asm_features_data).hexdigest(),
                length=5,
                tuf_version=5,
            ),
            "mock/ASM_DATA/1": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data1).hexdigest(),
                length=5,
                tuf_version=5,
            ),
            "mock/ASM_DATA/2": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data2).hexdigest(),
                length=5,
                tuf_version=5,
            ),
        }

        second_config = {
            "mock/ASM_FEATURES": ConfigMetadata(
                id="",
                product_name="ASM_FEATURES",
                sha256_hash=hashlib.sha256(asm_features_data).hexdigest(),
                length=5,
                tuf_version=5,
            ),
            "mock/ASM_DATA/2": ConfigMetadata(
                id="",
                product_name="ASM_DATA",
                sha256_hash=hashlib.sha256(asm_data_data_empty).hexdigest(),
                length=5,
                tuf_version=5,
            ),
        }

        RemoteConfig._worker._client._remove_previously_applied_configurations({}, first_config, {})

        RemoteConfig._worker._client._load_new_configurations(applied_configs, first_config, payload=payload)
        RemoteConfig._worker._client._applied_configs = {}

        mock_appsec_rules_data.assert_called_with(ANY, {"data": [{"a": 1}], "data2": [{"b": 2}]})
        mock_appsec_rules_data.reset_mock()

        payload = AgentPayload(
            target_files=[
                TargetFile(path="mock/ASM_FEATURES", raw=base64.b64encode(asm_features_data)),
                TargetFile(path="mock/ASM_DATA/1", raw=base64.b64encode(asm_data_data1)),
                TargetFile(path="mock/ASM_DATA/2", raw=base64.b64encode(asm_data_data_empty)),
            ]
        )

        RemoteConfig._worker._client._load_new_configurations({}, second_config, payload=payload)
        mock_appsec_rules_data.assert_called_with(ANY, {"data": [{"a": 1}], "data2": []})


def test_rc_activation_ip_blocking_data(tracer, remote_config_worker):
    with override_env({APPSEC_ENV: "true"}):
        tracer.configure(appsec_enabled=True, api_version="v0.4")
        rc_config = {
            "rules_data": [
                {
                    "data": [{"expiration": 1755346879, "value": "user8"}],
                    "id": "blocked_users",
                    "type": "data_with_expiration",
                },
                {
                    "data": [
                        {"value": "8.8.4.4"},
                    ],
                    "id": "blocked_ips",
                    "type": "ip_with_expiration",
                },
            ]
        }

        assert not RemoteConfig._worker

        RCAppSecCallBack(tracer)(None, rc_config)
        with _asm_request_context.asm_request_context_manager("8.8.4.4", {}):
            with tracer.trace("test", span_type=SpanTypes.WEB) as span:
                set_http_meta(
                    span,
                    Config(),
                )
            assert "triggers" in json.loads(span.get_tag(APPSEC_JSON))
            assert _context.get_item("http.request.remote_ip", span) == "8.8.4.4"


def test_rc_activation_ip_blocking_data_expired(tracer, remote_config_worker):
    with override_env({APPSEC_ENV: "true"}):
        tracer.configure(appsec_enabled=True, api_version="v0.4")
        rc_config = {
            "rules_data": [
                {
                    "data": [
                        {"expiration": int(time.time()) - 10000, "value": "8.8.4.4"},
                    ],
                    "id": "blocked_ips",
                    "type": "ip_with_expiration",
                },
            ]
        }

        assert not RemoteConfig._worker

        RCAppSecCallBack(tracer)(None, rc_config)

        with _asm_request_context.asm_request_context_manager("8.8.4.4", {}):
            with tracer.trace("test", span_type=SpanTypes.WEB) as span:
                set_http_meta(
                    span,
                    Config(),
                )
            assert span.get_tag(APPSEC_JSON) is None


def test_rc_activation_ip_blocking_data_not_expired(tracer, remote_config_worker):
    with override_env({APPSEC_ENV: "true"}):
        tracer.configure(appsec_enabled=True, api_version="v0.4")
        rc_config = {
            "rules_data": [
                {
                    "data": [
                        {"expiration": int(time.time()) + 10000, "value": "8.8.4.4"},
                    ],
                    "id": "blocked_ips",
                    "type": "ip_with_expiration",
                },
            ]
        }

        assert not RemoteConfig._worker

        RCAppSecCallBack(tracer)(None, rc_config)

        with _asm_request_context.asm_request_context_manager("8.8.4.4", {}):
            with tracer.trace("test", span_type=SpanTypes.WEB) as span:
                set_http_meta(
                    span,
                    Config(),
                )
            assert "triggers" in json.loads(span.get_tag(APPSEC_JSON))
            assert _context.get_item("http.request.remote_ip", span) == "8.8.4.4"


def test_rc_rules_data(tracer):
    RULES_PATH = os.path.join(os.path.dirname(os.path.dirname(ROOT_DIR)), "ddtrace/appsec/rules.json")
    with override_global_config(dict(_appsec_enabled=True)), override_env({APPSEC_ENV: "true"}), open(
        RULES_PATH, "r"
    ) as dd_rules:
        tracer.configure(appsec_enabled=True, api_version="v0.4")
        config = {
            "rules_data": [],
            "custom_rules": [],
            "rules": json.load(dd_rules)["rules"],
        }
        assert _appsec_rules_data(tracer, config)


def test_rc_rules_data_error_empty(tracer):
    with override_global_config(dict(_appsec_enabled=True)), override_env({APPSEC_ENV: "true"}):
        tracer.configure(appsec_enabled=True, api_version="v0.4")
        config = {}
        assert not _appsec_rules_data(tracer, config)


@pytest.mark.skipif(sys.version_info < (3, 5), reason="Python 2 is handling that test differently")
def test_rc_rules_data_error_ddwaf(tracer):
    with override_global_config(dict(_appsec_enabled=True)), override_env({APPSEC_ENV: "true"}):
        tracer.configure(appsec_enabled=True, api_version="v0.4")
        config = {
            "rules": [{"invalid": mock.MagicMock()}],
        }
        assert not _appsec_rules_data(tracer, config)



import abc
import json
import os
import sys
from threading import Thread
from types import FrameType
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Type
from typing import Union
from typing import cast

import six

from ddtrace.debugging._capture import utils
from ddtrace.debugging._capture.model import CapturedEvent
from ddtrace.debugging._capture.snapshot import Snapshot
from ddtrace.debugging._capture.snapshot import _capture_context
from ddtrace.debugging._config import config
from ddtrace.debugging._probe.model import CaptureLimits
from ddtrace.debugging._probe.model import FunctionLocationMixin
from ddtrace.debugging._probe.model import LineLocationMixin
from ddtrace.debugging._probe.model import LogProbeMixin
from ddtrace.internal import forksafe
from ddtrace.internal._encoding import BufferFull
from ddtrace.internal.logger import get_logger


log = get_logger(__name__)


class JsonBuffer(object):
    def __init__(self, max_size=None):
        self.max_size = max_size
        self._reset()

    def put(self, item):
        # type: (bytes) -> int
        if self._flushed:
            self._reset()

        size = len(item)
        if self.size + size > self.max_size:
            raise BufferFull(self.size, size)

        if self.size > 2:
            self.size += 1
            self._buffer += b","
        self._buffer += item
        self.size += size
        return size

    def _reset(self):
        self.size = 2
        self._buffer = bytearray(b"[")
        self._flushed = False

    def flush(self):
        self._buffer += b"]"
        try:
            return self._buffer
        finally:
            self._flushed = True


class Encoder(six.with_metaclass(abc.ABCMeta)):
    @abc.abstractmethod
    def encode(self, item):
        # type: (Any) -> bytes
        """Encode the given snapshot."""


class BufferedEncoder(six.with_metaclass(abc.ABCMeta)):
    count = 0

    @abc.abstractmethod
    def put(self, item):
        # type: (Any) -> int
        """Enqueue the given item and returns its encoded size."""

    @abc.abstractmethod
    def encode(self):
        # type: () -> Optional[bytes]
        """Encode the given item."""


_EMPTY_CAPTURED_CONTEXT = _capture_context([], [], (None, None, None), CaptureLimits())


def _probe_details(probe):
    # type (Probe) -> Dict[str, Any]
    if isinstance(probe, LineLocationMixin):
        location = {
            "file": probe.source_file,
            "lines": [probe.line],
        }
    elif isinstance(probe, FunctionLocationMixin):
        location = {
            "type": probe.module,
            "method": probe.func_qname,
        }
    return {
        "id": probe.probe_id,
        "location": location,
    }


def _snapshot_data(snapshot):
    # type (Snapshot) -> Dict[str, Any]
    frame = snapshot.frame
    probe = snapshot.probe

    captures = None
    if isinstance(probe, LogProbeMixin) and probe.take_snapshot:
        if isinstance(probe, LineLocationMixin):
            captures = {"lines": {probe.line: snapshot.line_capture or _EMPTY_CAPTURED_CONTEXT}}
        elif isinstance(probe, FunctionLocationMixin):
            captures = {
                "entry": snapshot.entry_capture or _EMPTY_CAPTURED_CONTEXT,
                "return": snapshot.return_capture or _EMPTY_CAPTURED_CONTEXT,
            }

    return {
        "id": snapshot.event_id,
        "timestamp": int(snapshot.timestamp * 1e3),  # milliseconds
        "duration": snapshot.duration,  # nanoseconds
        "stack": utils.capture_stack(frame),
        "evaluationErrors": [{"expr": e.expr, "message": e.message} for e in snapshot.errors],
        "captures": captures,
        "probe": _probe_details(snapshot.probe),
        "language": "python",
    }


def _logs_track_logger_details(thread, frame):
    # type: (Thread, FrameType) -> Dict[str, Any]
    code = frame.f_code

    return {
        "name": code.co_filename,
        "method": code.co_name,
        "thread_name": "%s;pid:%d" % (thread.name, os.getpid()),
        "thread_id": thread.ident,
        "version": 2,
    }


def add_tags(payload):
    if not config._tags_in_qs and config.tags:
        payload["ddtags"] = config.tags


def format_captured_value(value):
    # type: (Any) -> str
    v = value.get("value")
    if v is not None:
        return v
    elif value.get("isNull"):
        return "None"

    es = value.get("elements")
    if es is not None:
        return "%s(%s)" % (value["type"], ", ".join(format_captured_value(e) for e in es))

    es = value.get("entries")
    if es is not None:
        return "{%s}" % ", ".join(format_captured_value(k) + ": " + format_captured_value(v) for k, v in es)

    fs = value.get("fields")
    if fs is not None:
        return "%s(%s)" % (value["type"], ", ".join("%s=%s" % (k, format_captured_value(v)) for k, v in fs.items()))

    return "%s()" % value["type"]


def format_message(function, args, retval=None):
    # type: (str, Dict[str, Any], Optional[Any]) -> str
    message = "%s(%s)" % (
        function,
        ", ".join(("=".join((n, format_captured_value(a))) for n, a in args.items())),
    )

    if retval is not None:
        return "\n".join((message, "=".join(("@return", format_captured_value(retval)))))

    return message


def snapshot_message(snapshot, snapshot_data):
    # type: (Snapshot, Dict[str,Any]) -> str

    # if snapshot has generated message, return it
    if snapshot.message:
        return snapshot.message

    # create default
    top_frame = snapshot_data["stack"][0]
    if isinstance(snapshot.probe, LineLocationMixin):
        arguments = list(snapshot_data["captures"]["lines"].values())[0]["arguments"]
        return format_message(top_frame["function"], arguments)
    if isinstance(snapshot.probe, FunctionLocationMixin):
        arguments = snapshot_data["captures"]["entry"]["arguments"]
        retval = snapshot.return_capture["locals"].get("@return") if snapshot.return_capture else None
        return format_message(cast(str, snapshot.probe.func_qname), arguments, retval)
    return "snapshot " + snapshot.event_id


def _build_log_track_payload(
    service,  # type: str
    event,  # type: CapturedEvent
    message,  # type: str
    snapshot_data,  # type: Dict[str,Any]
    host,  # type: Optional[str]
):
    # type: (...) -> Dict[str, Any]
    context = event.context

    payload = {
        "service": service,
        "debugger.snapshot": snapshot_data,
        "host": host,
        "logger": _logs_track_logger_details(event.thread, event.frame),
        "dd.trace_id": context.trace_id if context else None,
        "dd.span_id": context.span_id if context else None,
        "ddsource": "dd_debugger",
        "message": message,
        "timestamp": int(event.timestamp * 1e3),  # milliseconds,
    }
    add_tags(payload)
    return payload


def logs_track_upload_snapshot_request(
    service,  # type: str
    snapshot,  # type: Snapshot
    host,  # type: Optional[str]
):
    # type: (...) -> Dict[str, Any]
    snapshot_data = _snapshot_data(snapshot)
    message = snapshot_message(snapshot, snapshot_data)

    return _build_log_track_payload(
        service=service,
        event=snapshot,
        message=message,
        snapshot_data=snapshot_data,
        host=host,
    )


class SnapshotJsonEncoder(Encoder):
    def __init__(self, service, host=None):
        # type: (str, Optional[str]) -> None
        self._service = service
        self._host = host

    def encode(self, snapshot):
        # type: (Snapshot) -> bytes
        return json.dumps(
            logs_track_upload_snapshot_request(
                service=self._service,
                snapshot=snapshot,
                host=self._host,
            )
        ).encode("utf-8")


class BatchJsonEncoder(BufferedEncoder):
    def __init__(self, item_encoders, buffer_size=4 * (1 << 20), on_full=None):
        # type: (Dict[Type, Union[Encoder, Type]], int, Optional[Callable[[Any, bytes], None]]) -> None
        self._encoders = item_encoders
        self._buffer = JsonBuffer(buffer_size)
        self._lock = forksafe.Lock()
        self._on_full = on_full
        self.count = 0
        self.max_size = buffer_size - self._buffer.size

    def put(self, item):
        # type: (Union[Snapshot, str]) -> int
        encoder = self._encoders.get(type(item))
        if encoder is None:
            raise ValueError("No encoder for item type: %r" % type(item))

        return self.put_encoded(item, encoder.encode(item))

    def put_encoded(self, item, encoded):
        # type: (Union[Snapshot, str], bytes) -> int
        try:
            with self._lock:
                size = self._buffer.put(encoded)
                self.count += 1
                return size
        except BufferFull:
            if self._on_full is not None:
                self._on_full(item, encoded)
            six.reraise(*sys.exc_info())

    def encode(self):
        # type: () -> Optional[bytes]
        with self._lock:
            if self.count == 0:
                # Reclaim memory
                self._buffer._reset()
                return None

            encoded = self._buffer.flush()
            self.count = 0
            return encoded



#!/usr/bin/env python3
import logging


logging.basicConfig(level=logging.DEBUG)

from tests.appsec.iast.fixtures.integration.print_str import print_str  # noqa: E402


def main():
    print_str()


if __name__ == "__main__":
    main()



import contextlib
from typing import TYPE_CHECKING

from ddtrace import config
from ddtrace.internal.logger import get_logger


if TYPE_CHECKING:
    from typing import Any
    from typing import Callable
    from typing import Generator
    from typing import List
    from typing import Optional
    from typing import Tuple

from ddtrace.vendor import contextvars


log = get_logger(__name__)

"""
Stopgap module for providing ASM context for the blocking features wrapping some
contextvars. When using this, note that context vars are always thread-local so each
thread will have a different context.
"""


# FIXME: remove these and use the new context API once implemented and allowing
# contexts without spans

_DD_EARLY_IP_CONTEXTVAR = contextvars.ContextVar("datadog_early_ip_contextvar", default=None)
_DD_EARLY_HEADERS_CONTEXTVAR = contextvars.ContextVar("datadog_early_headers_contextvar", default=None)
_DD_EARLY_HEADERS_CASE_SENSITIVE_CONTEXTVAR = contextvars.ContextVar(
    "datadog_early_headers_casesensitive_contextvar", default=False
)
_DD_BLOCK_REQUEST_CALLABLE = contextvars.ContextVar("datadog_block_request_callable_contextvar", default=None)
_DD_WAF_CALLBACK = contextvars.ContextVar("datadog_early_waf_callback", default=None)
_DD_WAF_RESULTS = contextvars.ContextVar("datadog_early_waf_results", default=([[], [], []]))


def reset():  # type: () -> None
    _DD_EARLY_IP_CONTEXTVAR.set(None)
    _DD_EARLY_HEADERS_CONTEXTVAR.set(None)
    _DD_EARLY_HEADERS_CASE_SENSITIVE_CONTEXTVAR.set(False)
    _DD_BLOCK_REQUEST_CALLABLE.set(None)


def set_ip(ip):  # type: (Optional[str]) -> None
    _DD_EARLY_IP_CONTEXTVAR.set(ip)


def get_ip():  # type: () -> Optional[str]
    return _DD_EARLY_IP_CONTEXTVAR.get()


# Note: get/set headers use Any since we just carry the headers here without changing or using them
# and different frameworks use different types that we don't want to force it into a Mapping at the
# early point set_headers is usually called


def set_headers(headers):  # type: (Any) -> None
    _DD_EARLY_HEADERS_CONTEXTVAR.set(headers)


def get_headers():  # type: () -> Optional[Any]
    return _DD_EARLY_HEADERS_CONTEXTVAR.get()


def set_headers_case_sensitive(case_sensitive):  # type: (bool) -> None
    _DD_EARLY_HEADERS_CASE_SENSITIVE_CONTEXTVAR.set(case_sensitive)


def get_headers_case_sensitive():  # type: () -> bool
    return _DD_EARLY_HEADERS_CASE_SENSITIVE_CONTEXTVAR.get()


def set_block_request_callable(_callable):  # type: (Optional[Callable]) -> None
    """
    Sets a callable that could be use to do a best-effort to block the request. If
    the callable need any params, like headers, they should be curried with
    functools.partial.
    """
    if _callable:
        _DD_BLOCK_REQUEST_CALLABLE.set(_callable)


def block_request():  # type: () -> None
    """
    Calls or returns the stored block request callable, if set.
    """
    _callable = _DD_BLOCK_REQUEST_CALLABLE.get()
    if _callable:
        _callable()

    log.debug("Block request called but block callable not set by framework")


def set_waf_callback(callback):  # type: (Any) -> None
    _DD_WAF_CALLBACK.set(callback)


def call_waf_callback(custom_data=None):
    # type: (dict[str, Any] | None) -> None
    if not config._appsec_enabled:
        return
    callback = _DD_WAF_CALLBACK.get()
    if callback:
        return callback(custom_data)
    else:
        log.warning("WAF callback called but not set")


def asm_request_context_set(remote_ip=None, headers=None, headers_case_sensitive=False, block_request_callable=None):
    # type: (Optional[str], Any, bool, Optional[Callable]) -> None
    set_ip(remote_ip)
    set_headers(headers)
    set_headers_case_sensitive(headers_case_sensitive)
    set_block_request_callable(block_request_callable)


def set_waf_results(result_data, result_info, is_blocked):  # type: (Any, Any, bool) -> None
    list_results_data, list_result_info, list_is_blocked = get_waf_results()
    list_results_data.append(result_data)
    list_result_info.append(result_info)
    list_is_blocked.append(is_blocked)
    _DD_WAF_RESULTS.set((list_results_data, list_result_info, list_is_blocked))


def get_waf_results():  # type: () -> Tuple[List[Any], List[Any], List[bool]]
    return _DD_WAF_RESULTS.get()


def reset_waf_results():  # type: () -> None
    _DD_WAF_RESULTS.set([[], [], []])


@contextlib.contextmanager
def asm_request_context_manager(
    remote_ip=None, headers=None, headers_case_sensitive=False, block_request_callable=None
):
    # type: (Optional[str], Any, bool, Optional[Callable]) -> Generator[None, None, None]
    asm_request_context_set(remote_ip, headers, headers_case_sensitive, block_request_callable)
    try:
        yield
    finally:
        reset()



#!/usr/bin/env python3
import os
import subprocess

import pytest
from six import PY2

from ddtrace.appsec.iast._util import _is_python_version_supported


def _run_python_file(*args, **kwargs):
    current_dir = os.path.dirname(__file__)
    cmd = [
        "ddtrace-run",
        "-d",
        "python",
        os.path.join(current_dir, "fixtures", "integration", kwargs.get("filename", "main.py")),
    ] + list(args)
    if "env" in kwargs:
        ret = subprocess.run(cmd, cwd=current_dir, env=kwargs["env"])
    else:
        ret = subprocess.run(cmd, cwd=current_dir)
    assert ret.returncode == 0


@pytest.mark.skipif(not _is_python_version_supported(), reason="IAST compatible versions")
def test_env_var_iast_enabled(capfd):
    # type: (...) -> None
    env = os.environ.copy()
    env["DD_IAST_ENABLED"] = "true"
    _run_python_file(env=env)
    captured = capfd.readouterr()
    assert "IAST enabled" in captured.err
    assert "hi" in captured.out


@pytest.mark.skipif(PY2, reason="Not testing Python 2")
def test_env_var_iast_disabled(monkeypatch, capfd):
    # type: (...) -> None
    env = os.environ.copy()
    env["DD_IAST_ENABLED"] = "false"
    _run_python_file(env=env)
    captured = capfd.readouterr()
    assert "hi" in captured.out
    assert "IAST enabled" not in captured.err


@pytest.mark.skipif(PY2, reason="Not testing Python 2")
def test_env_var_iast_unset(monkeypatch, capfd):
    # type: (...) -> None
    _run_python_file()
    captured = capfd.readouterr()
    assert "hi" in captured.out
    assert "IAST enabled" not in captured.err


@pytest.mark.skipif(not _is_python_version_supported(), reason="IAST compatible versions")
@pytest.mark.xfail(reason="IAST now working with Gevent yet")
def test_env_var_iast_enabled_gevent_unload_modules_true(capfd):
    # type: (...) -> None
    env = os.environ.copy()
    env["DD_IAST_ENABLED"] = "true"
    env["DD_UNLOAD_MODULES_FROM_SITECUSTOMIZE"] = "true"
    _run_python_file(filename="main_gevent.py", env=env)
    captured = capfd.readouterr()
    assert "IAST enabled" in captured.err
    assert "hi" in captured.out


@pytest.mark.skipif(not _is_python_version_supported(), reason="IAST compatible versions")
@pytest.mark.xfail(reason="IAST now working with Gevent yet")
def test_env_var_iast_enabled_gevent_unload_modules_false(capfd):
    # type: (...) -> None
    env = os.environ.copy()
    env["DD_IAST_ENABLED"] = "true"
    env["DD_UNLOAD_MODULES_FROM_SITECUSTOMIZE"] = "false"
    _run_python_file(filename="main_gevent.py", env=env)
    captured = capfd.readouterr()
    assert "IAST enabled" in captured.err
    assert "hi" in captured.out


@pytest.mark.skipif(not _is_python_version_supported(), reason="IAST compatible versions")
@pytest.mark.xfail(reason="IAST now working with Gevent yet")
def test_env_var_iast_enabled_gevent_patch_all_true(capfd):
    # type: (...) -> None
    env = os.environ.copy()
    env["DD_IAST_ENABLED"] = "true"
    env["DD_GEVENT_PATCH_ALL"] = "true"
    _run_python_file(filename="main_gevent.py", env=env)
    captured = capfd.readouterr()
    assert "IAST enabled" in captured.err
    assert "hi" in captured.out



import json

import pytest

from ddtrace._monkey import IAST_PATCH
from ddtrace._monkey import patch_iast
from ddtrace.constants import IAST_CONTEXT_KEY
from ddtrace.constants import IAST_JSON
from ddtrace.constants import SAMPLING_PRIORITY_KEY
from ddtrace.constants import USER_KEEP
from ddtrace.ext import SpanTypes
from ddtrace.internal import _context
from tests.utils import DummyTracer
from tests.utils import override_env


def traced_function(tracer):
    with tracer.trace("test", span_type=SpanTypes.WEB) as span:
        import hashlib

        m = hashlib.new("md5")
        m.update(b"Nobody inspects")
        m.update(b" the spammish repetition")
        num_vulnerabilities = 10
        for i in range(0, num_vulnerabilities):
            m.digest()
    return span


def test_appsec_iast_processor():
    with override_env(dict(DD_IAST_ENABLED="true")):
        patch_iast(**IAST_PATCH)

        tracer = DummyTracer(iast_enabled=True)

        span = traced_function(tracer)
        tracer._on_span_finish(span)

        span_report = _context.get_item(IAST_CONTEXT_KEY, span=span)
        result = span.get_tag(IAST_JSON)

        assert len(span_report.vulnerabilities) == 1
        assert len(json.loads(result)["vulnerabilities"]) == 1


@pytest.mark.parametrize("sampling_rate", ["0.0", "0.5", "1.0"])
def test_appsec_iast_processor_ensure_span_is_manual_keep(sampling_rate):
    with override_env(dict(DD_IAST_ENABLED="true", DD_TRACE_SAMPLE_RATE=sampling_rate)):
        patch_iast(**IAST_PATCH)

        tracer = DummyTracer(iast_enabled=True)

        span = traced_function(tracer)
        tracer._on_span_finish(span)

        result = span.get_tag(IAST_JSON)

        assert len(json.loads(result)["vulnerabilities"]) == 1
        assert span.get_metric(SAMPLING_PRIORITY_KEY) is USER_KEEP



import os
import platform

import mock
import six

from ddtrace import ext
from ddtrace.profiling.collector import _lock
from ddtrace.profiling.collector import memalloc
from ddtrace.profiling.collector import stack_event
from ddtrace.profiling.exporter import _packages
from ddtrace.profiling.exporter import pprof


TEST_EVENTS = {
    stack_event.StackExceptionSampleEvent: [
        stack_event.StackExceptionSampleEvent(
            timestamp=1,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", "class"),
                ("foobar.py", 44, "func2", ""),
                ("foobar.py", 19, "func5", ""),
            ],
            exc_type=TypeError,
            sampling_period=1000000,
            nframes=3,
        ),
        stack_event.StackExceptionSampleEvent(
            timestamp=2,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", "class"),
                ("foobar.py", 44, "func2", ""),
                ("foobar.py", 20, "func5", "otherclass"),
            ],
            sampling_period=1000000,
            exc_type=TypeError,
            nframes=3,
        ),
        stack_event.StackExceptionSampleEvent(
            timestamp=3,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar.py", 19, "func5", ""),
            ],
            sampling_period=1000000,
            exc_type=TypeError,
            nframes=4,
        ),
        stack_event.StackExceptionSampleEvent(
            timestamp=4,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", "class"),
                ("foobar2.py", 19, "func5", ""),
            ],
            sampling_period=1000000,
            exc_type=TypeError,
            nframes=6,
        ),
        stack_event.StackExceptionSampleEvent(
            timestamp=5,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", "classX"),
                ("foobar2.py", 19, "func5", ""),
            ],
            sampling_period=1000000,
            exc_type=ValueError,
            nframes=3,
        ),
        stack_event.StackExceptionSampleEvent(
            timestamp=6,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar.py", 19, "func5", ""),
            ],
            sampling_period=1000000,
            exc_type=IOError,
            nframes=3,
        ),
        stack_event.StackExceptionSampleEvent(
            timestamp=7,
            thread_id=67892304,
            thread_native_id=123987,
            local_root_span_id=1322219321,
            span_id=1322219,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 49, "func2", ""),
                ("foobar.py", 19, "func5", ""),
            ],
            sampling_period=1000000,
            exc_type=IOError,
            nframes=3,
        ),
    ],
    memalloc.MemoryAllocSampleEvent: [
        memalloc.MemoryAllocSampleEvent(
            timestamp=1,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar.py", 19, "func5", ""),
            ],
            size=34,
            capture_pct=44,
            nframes=3,
            nevents=1024,
        ),
        memalloc.MemoryAllocSampleEvent(
            timestamp=2,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar.py", 20, "func5", ""),
            ],
            size=99,
            capture_pct=100,
            nframes=3,
            nevents=1024,
        ),
        memalloc.MemoryAllocSampleEvent(
            timestamp=3,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar.py", 19, "func5", ""),
            ],
            size=340,
            capture_pct=50,
            nframes=4,
            nevents=1024,
        ),
        memalloc.MemoryAllocSampleEvent(
            timestamp=4,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar2.py", 19, "func5", ""),
            ],
            size=44,
            capture_pct=33,
            nframes=6,
            nevents=1024,
        ),
        memalloc.MemoryAllocSampleEvent(
            timestamp=5,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            local_root_span_id=1322219321,
            span_id=49393,
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar2.py", 19, "func5", ""),
            ],
            size=68,
            capture_pct=50,
            nframes=3,
            nevents=2048,
        ),
        memalloc.MemoryAllocSampleEvent(
            timestamp=6,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar.py", 19, "func5", ""),
            ],
            size=24,
            capture_pct=90,
            nframes=3,
            nevents=2048,
        ),
        memalloc.MemoryAllocSampleEvent(
            timestamp=7,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 49, "func2", ""),
                ("foobar.py", 19, "func5", ""),
            ],
            size=12,
            capture_pct=100,
            nframes=3,
            nevents=2048,
        ),
    ],
    memalloc.MemoryHeapSampleEvent: [
        memalloc.MemoryHeapSampleEvent(
            timestamp=1,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar.py", 19, "func5", ""),
            ],
            size=34,
            nframes=3,
            sample_size=512 * 1024,
        ),
        memalloc.MemoryHeapSampleEvent(
            timestamp=2,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar.py", 20, "func5", ""),
            ],
            size=99,
            nframes=3,
            sample_size=512 * 1024,
        ),
        memalloc.MemoryHeapSampleEvent(
            timestamp=3,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar.py", 19, "func5", ""),
            ],
            size=340,
            nframes=4,
            sample_size=512 * 1024,
        ),
        memalloc.MemoryHeapSampleEvent(
            timestamp=4,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar2.py", 19, "func5", ""),
            ],
            size=44,
            nframes=6,
            sample_size=512 * 1024,
        ),
        memalloc.MemoryHeapSampleEvent(
            timestamp=5,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            local_root_span_id=1322219321,
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar2.py", 19, "func5", ""),
            ],
            size=68,
            nframes=3,
            sample_size=512 * 1024,
        ),
        memalloc.MemoryHeapSampleEvent(
            timestamp=6,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar.py", 19, "func5", ""),
            ],
            size=24,
            nframes=3,
            sample_size=512 * 1024,
        ),
        memalloc.MemoryHeapSampleEvent(
            timestamp=7,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 49, "func2", ""),
                ("foobar.py", 19, "func5", ""),
            ],
            size=12,
            nframes=3,
            sample_size=512 * 1024,
        ),
    ],
    stack_event.StackSampleEvent: [
        stack_event.StackSampleEvent(
            timestamp=1,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            local_root_span_id=1322219321,
            span_id=49343,
            trace_type=ext.SpanTypes.WEB,
            trace_resource_container=["myresource"],
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar.py", 19, "func5", ""),
            ],
            task_id=123,
            task_name="sometask",
            wall_time_ns=1324,
            cpu_time_ns=1321,
            sampling_period=1000000,
            nframes=3,
        ),
        stack_event.StackSampleEvent(
            timestamp=2,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            local_root_span_id=1322219321,
            span_id=24930,
            trace_type="sql",
            trace_resource_container=["\x1bnotme"],
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar.py", 20, "func5", ""),
            ],
            wall_time_ns=13244,
            cpu_time_ns=1312,
            sampling_period=1000000,
            nframes=3,
        ),
        stack_event.StackSampleEvent(
            timestamp=3,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            local_root_span_id=1322219321,
            span_id=24930,
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar.py", 19, "func5", ""),
            ],
            wall_time_ns=1324,
            cpu_time_ns=29121,
            sampling_period=1000000,
            nframes=4,
        ),
        stack_event.StackSampleEvent(
            timestamp=4,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar2.py", 19, "func5", ""),
            ],
            wall_time_ns=213244,
            cpu_time_ns=94021,
            sampling_period=1000000,
            nframes=6,
        ),
        stack_event.StackSampleEvent(
            timestamp=5,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            local_root_span_id=1322219321,
            span_id=249304,
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar2.py", 19, "func5", ""),
            ],
            wall_time_ns=132444,
            cpu_time_ns=9042,
            sampling_period=1000000,
            nframes=3,
        ),
        stack_event.StackSampleEvent(
            timestamp=6,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar.py", 19, "func5", ""),
            ],
            wall_time_ns=18244,
            cpu_time_ns=841019,
            sampling_period=1000000,
            nframes=3,
        ),
        stack_event.StackSampleEvent(
            timestamp=7,
            thread_id=67892304,
            thread_native_id=123987,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 49, "func2", ""),
                ("foobar.py", 19, "func5", ""),
            ],
            wall_time_ns=13244,
            cpu_time_ns=501809,
            sampling_period=1000000,
            nframes=3,
        ),
    ],
    _lock.LockAcquireEvent: [
        _lock.LockAcquireEvent(
            lock_name="foobar.py:12",
            timestamp=1,
            thread_id=67892304,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar.py", 19, "func5", ""),
            ],
            task_id=12234,
            task_name="mytask",
            local_root_span_id=23435,
            span_id=345432,
            trace_type=ext.SpanTypes.WEB,
            trace_resource_container=["myresource"],
            nframes=3,
            wait_time_ns=74839,
            sampling_pct=10,
        ),
        _lock.LockAcquireEvent(
            lock_name="foobar.py:12",
            timestamp=2,
            thread_id=67892304,
            thread_name="MainThread",
            local_root_span_id=23435,
            span_id=345432,
            trace_type="sql",
            trace_resource_container=[b"notme"],
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar.py", 20, "func5", ""),
            ],
            nframes=3,
            wait_time_ns=7483,
            sampling_pct=10,
        ),
        _lock.LockAcquireEvent(
            lock_name="foobar.py:12",
            timestamp=3,
            thread_id=67892304,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar.py", 19, "func5", ""),
            ],
            nframes=4,
            wait_time_ns=7489,
            sampling_pct=10,
        ),
        _lock.LockAcquireEvent(
            lock_name="foobar.py:12",
            timestamp=4,
            thread_id=67892304,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar2.py", 19, "func5", ""),
            ],
            nframes=6,
            wait_time_ns=4839,
            sampling_pct=10,
        ),
        _lock.LockAcquireEvent(
            lock_name="foobar.py:12",
            timestamp=5,
            thread_id=67892304,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar2.py", 19, "func5", ""),
            ],
            nframes=3,
            wait_time_ns=748394,
            sampling_pct=10,
        ),
        _lock.LockAcquireEvent(
            lock_name="foobar.py:12",
            timestamp=6,
            thread_id=67892304,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar.py", 19, "func5", ""),
            ],
            nframes=3,
            wait_time_ns=748339,
            sampling_pct=10,
        ),
        _lock.LockAcquireEvent(
            lock_name="foobar.py:12",
            timestamp=7,
            thread_id=67892304,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 49, "func2", ""),
                ("foobar.py", 19, "func5", ""),
            ],
            nframes=3,
            wait_time_ns=174839,
            sampling_pct=10,
        ),
    ],
    _lock.LockReleaseEvent: [
        _lock.LockReleaseEvent(
            lock_name="foobar.py:12",
            timestamp=1,
            thread_id=67892304,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar.py", 19, "func5", ""),
            ],
            nframes=3,
            locked_for_ns=74839,
            sampling_pct=5,
        ),
        _lock.LockReleaseEvent(
            lock_name="foobar.py:12",
            timestamp=2,
            thread_id=67892304,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar.py", 20, "func5", ""),
            ],
            nframes=3,
            locked_for_ns=7483,
            sampling_pct=5,
        ),
        _lock.LockReleaseEvent(
            lock_name="foobar.py:12",
            timestamp=3,
            thread_id=67892304,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar.py", 19, "func5", ""),
            ],
            nframes=4,
            locked_for_ns=7489,
            sampling_pct=5,
        ),
        _lock.LockReleaseEvent(
            lock_name="foobar.py:12",
            timestamp=4,
            thread_id=67892304,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar2.py", 19, "func5", ""),
            ],
            nframes=6,
            locked_for_ns=4839,
            sampling_pct=5,
        ),
        _lock.LockReleaseEvent(
            lock_name="foobar.py:12",
            timestamp=5,
            thread_id=67892304,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar2.py", 19, "func5", ""),
            ],
            nframes=3,
            locked_for_ns=748394,
            sampling_pct=5,
        ),
        _lock.LockReleaseEvent(
            lock_name="foobar.py:12",
            timestamp=6,
            thread_id=67892304,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 44, "func2", ""),
                ("foobar.py", 19, "func5", ""),
            ],
            nframes=3,
            locked_for_ns=748339,
            sampling_pct=5,
        ),
        _lock.LockReleaseEvent(
            lock_name="foobar.py:12",
            timestamp=7,
            thread_id=67892304,
            thread_name="MainThread",
            frames=[
                ("foobar.py", 23, "func1", ""),
                ("foobar.py", 49, "func2", ""),
                ("foobar.py", 19, "func5", ""),
            ],
            nframes=3,
            locked_for_ns=174839,
            sampling_pct=50,
        ),
    ],
}


def test_string_table():
    t = pprof._StringTable()
    assert len(t) == 1
    id1 = t.index("foobar")
    assert len(t) == 2
    assert id1 == t.index("foobar")
    assert len(t) == 2
    id2 = t.index("foobaz")
    assert len(t) == 3
    assert id2 == t.index("foobaz")
    assert len(t) == 3
    assert id1 != id2


def test_to_str_none():
    c = pprof._PprofConverter()
    id1 = c._str(None)
    id2 = c._str("None")
    id_o = c._str("foobar")
    assert id1 == id2 != id_o


@mock.patch("ddtrace.internal.utils.config.get_application_name")
def test_pprof_exporter(gan):
    gan.return_value = "bonjour"
    exp = pprof.PprofExporter()
    exports, _ = exp.export(TEST_EVENTS, 1, 7)

    assert len(exports.sample_type) == 11
    assert len(exports.string_table) == 58
    assert len(exports.sample) == 22 if six.PY2 else 28
    assert len(exports.location) == 8

    assert exports.period == 1000000
    assert exports.time_nanos == 1
    assert exports.duration_nanos == 6

    assert all(_ in exports.string_table for _ in ("time", "nanoseconds", "bonjour"))


@mock.patch("ddtrace.internal.utils.config.get_application_name")
def test_pprof_exporter_libs(gan):
    _packages._FILE_PACKAGE_MAPPING = _packages._build_package_file_mapping()
    gan.return_value = "bonjour"
    exp = pprof.PprofExporter()
    TEST_EVENTS = {
        stack_event.StackSampleEvent: [
            stack_event.StackSampleEvent(
                timestamp=1,
                thread_id=67892304,
                thread_native_id=123987,
                thread_name="MainThread",
                local_root_span_id=1322219321,
                span_id=49343,
                trace_type=ext.SpanTypes.WEB,
                trace_resource_container=["myresource"],
                frames=[
                    (six.__file__, 23, "func1", ""),
                    (memalloc.__file__, 44, "func2", ""),
                    ("foobar.py", 19, "func5", ""),
                ],
                task_id=123,
                task_name="sometask",
                wall_time_ns=1324,
                cpu_time_ns=1321,
                sampling_period=1000000,
                nframes=3,
            ),
            stack_event.StackSampleEvent(
                timestamp=2,
                thread_id=67892304,
                thread_native_id=123987,
                thread_name="MainThread",
                local_root_span_id=1322219321,
                span_id=24930,
                trace_type="sql",
                trace_resource_container=["\x1bnotme"],
                frames=[
                    (__file__, 23, "func1", ""),
                    ("foobar.py", 44, "func2", ""),
                    (os.__file__, 20, "func5", ""),
                ],
                wall_time_ns=13244,
                cpu_time_ns=1312,
                sampling_period=1000000,
                nframes=3,
            ),
        ]
    }

    _, libs = exp.export(TEST_EVENTS, 1, 7)

    # Version does not match between pip and __version__ for ddtrace; ignore
    for lib in libs:
        if lib["name"] == "ddtrace":
            del lib["version"]
        elif lib["kind"] == "standard library":
            assert len(lib["paths"]) == 1
            del lib["paths"]
        elif lib["name"] == "<unknown>":
            assert len(lib["paths"]) == 1
            del lib["paths"]

    expected_libs = [
        {"name": "ddtrace", "kind": "library", "paths": {__file__, memalloc.__file__}},
        {"name": "six", "kind": "library", "version": six.__version__, "paths": {six.__file__}},
        {"kind": "standard library", "name": "stdlib", "version": platform.python_version()},
        {
            "kind": "library",
            "name": "<unknown>",
            "version": "<unknown>",
        },
    ]

    if six.PY3:
        expected_libs.append(
            {"kind": "standard library", "name": "platstdlib", "version": platform.python_version()},
        )

    # DEV: We cannot convert the lists to sets because the some of the values
    # in the dicts are list. We resort to matching the elements of one list to
    # the other instead and check that:
    # - for all elements in libs we have a match in expected_libs
    # - we end up with an empty expected_libs
    # This is equivalent to checking that the two lists are equal.
    for _ in libs:
        if "paths" in _:
            _["paths"] = set(_["paths"])
        expected_libs.remove(_)

    assert not expected_libs


def test_pprof_exporter_empty():
    exp = pprof.PprofExporter()
    export, libs = exp.export({}, 0, 1)
    assert len(libs) > 0
    assert len(export.sample) == 0



"""IAST (interactive application security testing) analyzes code for security vulnerabilities.

To add new vulnerabilities analyzers (Taint sink) we should update `IAST_PATCH` in `ddtrace._monkey.patch_iast` function
`IAST_PATCH['[my_new_vulnerability]'] = True`

Create new file with the same name: `ddtrace/appsec/iast/taint_sinks/[my_new_vulnerability].py`

Then, implement the `patch()` function and its wrappers.

In order to have the better performance, the Overhead control engine (OCE) helps us to control the overhead of our
wrapped functions. We should create a class that inherit from `ddtrace.appsec.iast.taint_sinks._base.VulnerabilityBase`
and register with `ddtrace.appsec.iast.oce`.

@oce.register
class MyVulnerability(VulnerabilityBase):
    vulnerability_type = "MyVulnerability"
    evidence_type = "kind_of_Vulnerability"

Before that, we should decorate our wrappers with `wrap` method and
report the vulnerabilities with `report` method. OCE will manage the number of requests, number of vulnerabilities
to reduce the overhead.

@WeakHash.wrap
def wrapped_function(wrapped, instance, args, kwargs):
    # type: (Callable, str, Any, Any, Any) -> Any
    WeakHash.report(
        evidence_value=evidence,
    )
    return wrapped(*args, **kwargs)
"""  # noqa: RST201, RST213, RST210
from ddtrace.appsec.iast._overhead_control_engine import OverheadControl


oce = OverheadControl()


__all__ = [
    "oce",
]
